# Copied from Megatron-LM: https://github.com/NVIDIA/Megatron-LM
# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.

from typing import Callable, List, Optional, Tuple, Union
import itertools
import os
import warnings

from torch.cuda.amp import custom_bwd, custom_fwd
from torch.nn.parameter import Parameter
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init

from realhf.base import constants

from .mappings import (
    copy_to_tensor_model_parallel_region,
    gather_from_sequence_parallel_region,
    gather_from_tensor_model_parallel_region,
    reduce_from_tensor_model_parallel_region,
    reduce_scatter_to_sequence_parallel_region,
    scatter_to_tensor_model_parallel_region,
)
from .utils import (
    _initialize_affine_weight_gpu,
    divide,
    set_tensor_model_parallel_attributes,
    VocabUtility,
)

_grad_accum_fusion_available = True
try:
    import fused_weight_gradient_mlp_cuda
except ImportError:
    _grad_accum_fusion_available = False

import realhf.base.logging as logging

logger = logging.getLogger("model_parallel.modules")


def get_activation_fn(activation_function: str) -> Callable:
    if activation_function == "gelu":
        return nn.functional.gelu
    elif activation_function == "gelu_new":
        from realhf.impl.model.modules.activations import new_gelu_activation

        return new_gelu_activation
    elif activation_function == "silu":
        return nn.SiLU()
    else:
        raise NotImplementedError(
            'Only "gelu" activation function is available.'
        )


class ParallelEmbedding(torch.nn.Module):
    """Embedding parallelized in the vocabulary dimension.

    This is mainly adapted from torch.nn.Embedding and all the default
    values are kept.
    Arguments:
        num_embeddings: vocabulary size.
        embedding_dim: size of hidden state.

    Keyword Arguments:
        init_method: method to initialize weights.
        perform_initialization
    """

    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        init_method=init.xavier_normal_,
        # params_dtype: torch.dtype=torch.float32,
        perform_initialization: bool = True,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super(ParallelEmbedding, self).__init__()
        # Keep the input dimensions.
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        # Set the detauls for compatibility.
        self.padding_idx = None
        self.max_norm = None
        self.norm_type = 2.0
        self.scale_grad_by_freq = False
        self.sparse = False
        self._weight = None
        self.tensor_model_parallel_size = constants.model_parallel_world_size()
        # Divide the weight matrix along the vocaburaly dimension.
        self.vocab_start_index, self.vocab_end_index = (
            VocabUtility.vocab_range_from_global_vocab_size(
                self.num_embeddings,
                constants.model_parallel_rank(),
                self.tensor_model_parallel_size,
            )
        )
        self.num_embeddings_per_partition = (
            self.vocab_end_index - self.vocab_start_index
        )

        logger.debug(
            f"ParallelEmbedding: num_embeddings={num_embeddings}, per_partition={self.num_embeddings_per_partition}, embedding_dim={embedding_dim},"
            f"tp_rank={constants.model_parallel_rank()},tp_world_size={constants.model_parallel_world_size()}"
        )
        # Allocate weights and initialize.
        self.weight = Parameter(
            torch.empty(
                self.num_embeddings_per_partition,
                self.embedding_dim,
                device=device,
                dtype=dtype,
            )
        )
        if perform_initialization:
            _initialize_affine_weight_gpu(
                self.weight, init_method, partition_dim=0, stride=1
            )

    def forward(self, input_) -> torch.Tensor:
        if self.tensor_model_parallel_size > 1:
            # Build the mask.
            input_mask = (input_ < self.vocab_start_index) | (
                input_ >= self.vocab_end_index
            )
            # Mask the input.
            masked_input = input_.clone() - self.vocab_start_index
            masked_input[input_mask] = 0
        else:
            masked_input = input_
            # Get the embeddings.

        output_parallel = F.embedding(
            masked_input,
            self.weight,
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )
        # Mask the output embedding.
        if self.tensor_model_parallel_size > 1:
            output_parallel[input_mask, :] = 0.0
        # Reduce across all the model parallel GPUs.
        output = reduce_from_tensor_model_parallel_region(output_parallel)
        return output


class LinearWithFrozenWeight(torch.autograd.Function):
    """Linear operator that does not calculate gradient for weight.
    This op and LinearWithGradAccumulationAndAsyncCommunication performs
    mathematically-identical forward and DGRAD.

    Conceptually this op is the same as torch.nn.functional.linear with
    weight.requires_grad==False, but in realhf.experiments they are not identical
    mathematically."""

    @staticmethod
    @custom_fwd
    def forward(
        ctx,
        input,
        weight,
        bias,
    ):
        ctx.save_for_backward(weight)
        output = torch.matmul(input, weight.t())
        if bias is not None:
            output = output + bias
        return output

    @staticmethod
    @custom_bwd
    def backward(ctx, grad_output):
        (weight,) = ctx.saved_tensors
        grad_input = grad_output.matmul(weight)
        return grad_input, None, None


def linear_with_frozen_weight(
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: Optional[torch.Tensor],
    gradient_accumulation_fusion: bool,
    async_grad_allreduce: bool,
    sequence_parallel: bool,
) -> torch.Tensor:
    """Linear layer execution with weight.requires_grad == False.

    This function handles linear layers with weight frozen (untrainable).
    In the forward, it only saves weight and does not save input activations.
    In the backward, it does not perform weight gradient calculation, or
    weight gradient allreduce.

    Arguments:

    input (torch.Tensor required): input like torch.nn.functional.linear

    weight (torch.Tensor required): weight like torch.nn.functional.linear

    bias (torch.Tensor optional): bias like torch.nn.functional.linear

    gradient_accumulation_fusion (bool required): dummy argument, used to
    keep the API unified between all forward implementation functions.

    async_grad_allreduce (bool required): dummy argument, used to
    keep the API unified between all forward implementation functions.

    sequence_parallel (bool required): Indicates that sequence
        parallelism is used and thus in the forward pass the input is
        all gathered, and the backward pass the input gradients are
        reduce scattered.
    """

    if sequence_parallel:
        input = gather_from_sequence_parallel_region(
            input, model_parallel_output_grad=True
        )
    else:
        input = input

    args = [
        input,
        weight,
        bias,
    ]

    return LinearWithFrozenWeight.apply(*args)


class LinearWithGradAccumulationAndAsyncCommunication(torch.autograd.Function):
    """See linear_with_grad_accumulation_and_async_allreduce"""

    @staticmethod
    @custom_fwd
    def forward(
        ctx,
        input,
        weight,
        bias,
        gradient_accumulation_fusion,
        async_grad_allreduce,
        sequence_parallel,
    ):
        # disable sequence parallel for now for it requires a global buffer
        ctx.save_for_backward(input, weight)
        ctx.use_bias = bias is not None
        ctx.gradient_accumulation_fusion = gradient_accumulation_fusion
        ctx.async_grad_allreduce = async_grad_allreduce
        ctx.sequence_parallel = sequence_parallel

        if sequence_parallel:
            assert (
                not ctx.async_grad_allreduce
            ), "async_grad_allreduce and sequence_parallel can not be both True"
            world_size = constants.model_parallel_world_size()
            dim_size = list(input.size())
            dim_size[0] = dim_size[0] * world_size

            all_gather_buffer = constants.get_global_memory_buffer().get_tensor(
                dim_size, input.dtype, "mpu"
            )
            torch.distributed._all_gather_base(
                all_gather_buffer, input, group=constants.model_parallel_group()
            )
            total_input = all_gather_buffer
        else:
            total_input = input

        output = torch.matmul(total_input, weight.t())
        if bias is not None:
            output = output + bias
        return output

    @staticmethod
    @custom_bwd
    def backward(ctx, grad_output):
        input, weight = ctx.saved_tensors
        use_bias = ctx.use_bias

        if ctx.sequence_parallel:
            world_size = constants.model_parallel_world_size()
            dim_size = list(input.size())
            dim_size[0] = dim_size[0] * world_size

            all_gather_buffer = constants.get_global_memory_buffer().get_tensor(
                dim_size, input.dtype, "mpu"
            )
            handle = torch.distributed._all_gather_base(
                all_gather_buffer,
                input,
                group=constants.model_parallel_group(),
                async_op=True,
            )

            # Here we rely on CUDA_DEVICE_MAX_CONNECTIONS=1 to ensure that the
            # gather is scheduled before the input gradient computation
            total_input = all_gather_buffer
        else:
            total_input = input
        grad_input = grad_output.matmul(weight)

        if ctx.sequence_parallel:
            handle.wait()

        # Doing gather + slicing during the NeMo forward pass can make this tensor
        # not be contiguous. PyTorch only checks if the tensor is contiguous, and only
        # clones it if it's not contiguous:
        # https://github.com/pytorch/pytorch/blob/c47cf9bc7f9e02f649ab4ed53fe4d35732c92ab6/torch/_refs/__init__.py#L2761
        grad_output = grad_output.contiguous()
        # Convert the tensor shapes to 2D for execution compatibility
        grad_output = grad_output.view(-1, grad_output.shape[-1])
        total_input = total_input.view(-1, total_input.shape[-1])

        if ctx.async_grad_allreduce:
            # Asynchronous all-reduce
            handle = torch.distributed.all_reduce(
                grad_input,
                group=constants.model_parallel_group(),
                async_op=True,
            )
            # Here we rely on CUDA_DEVICE_MAX_CONNECTIONS=1 to ensure that the
            # all-reduce is scheduled before the weight gradient computation

        if ctx.sequence_parallel:
            assert not ctx.async_grad_allreduce
            dim_size = list(input.size())
            sub_grad_input = torch.empty(
                dim_size,
                dtype=input.dtype,
                device=torch.cuda.current_device(),
                requires_grad=False,
            )
            # reduce_scatter
            handle = torch.distributed._reduce_scatter_base(
                sub_grad_input,
                grad_input,
                group=constants.model_parallel_group(),
                async_op=True,
            )
            # Here we rely on CUDA_DEVICE_MAX_CONNECTIONS=1 to ensure that the
            # reduce scatter is scheduled before the weight gradient computation

        if ctx.gradient_accumulation_fusion:
            if weight.main_grad.dtype == torch.float32:
                fused_weight_gradient_mlp_cuda.wgrad_gemm_accum_fp32(
                    total_input, grad_output, weight.main_grad
                )
            elif weight.main_grad.dtype in (torch.float16, torch.bfloat16):
                fused_weight_gradient_mlp_cuda.wgrad_gemm_accum_fp16(
                    total_input, grad_output, weight.main_grad
                )
            else:
                raise RuntimeError(
                    "Unsupported gradient type for gradient accumulation fusion"
                )

            if hasattr(weight, "grad_added_to_main_grad"):
                # When overlap_grad_reduce is True, need to ensure that backward hooks
                # are all run on the main backprop thread to prevent deadlocks. Setup
                # dummy grad_weight tensor to prevent backward hooks from being run
                # in a background thread.
                if getattr(weight, "zero_out_wgrad", False):
                    grad_weight = torch.zeros(
                        weight.main_grad.shape,
                        dtype=input.dtype,
                        device=torch.cuda.current_device(),
                        requires_grad=False,
                    )
                else:
                    grad_weight = torch.empty(
                        weight.main_grad.shape,
                        dtype=input.dtype,
                        device=torch.cuda.current_device(),
                        requires_grad=False,
                    )
                weight.grad_added_to_main_grad = True
            else:
                grad_weight = None
        else:
            grad_weight = grad_output.t().matmul(total_input)
        grad_bias = grad_output.sum(dim=0) if use_bias else None

        if ctx.sequence_parallel:
            handle.wait()
            return sub_grad_input, grad_weight, grad_bias, None, None, None

        if ctx.async_grad_allreduce:
            handle.wait()

        return grad_input, grad_weight, grad_bias, None, None, None


def linear_with_grad_accumulation_and_async_allreduce(
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: Optional[torch.Tensor],
    gradient_accumulation_fusion: bool,
    async_grad_allreduce: bool,
    sequence_parallel: bool,
) -> torch.Tensor:
    """Linear layer execution with asynchronous communication and
    gradient accumulation fusion in backprop.

    This has the option to accumulate the result of backprop
    calculation into an existing gradient buffer, preventing the need
    to do an additional addition kernel after the gradient
    calculation.

    Additionally, the tensor parallel all reduce of the input
    gradients can be done asynchronously with the calculation of
    the weight gradients.

    In the case of sequence parallelism, the reduce scatter of the
    input gradients is done asynchronously with the calcluation of the
    weight gradients.

    Use of this module requires that the environment variable
    CUDA_DEVICE_MAX_CONNECTIONS=1. There are a few collective
    operations, noted in the code, that should be scheduled before
    compute kernels to overlap the communication with the computation,
    which is necessary for a speedup but not for correctness so that
    ordering isn't imposed by the scheduler. Setting
    CUDA_DEVICE_MAX_CONNECTIONS=1 forces the kernels to be scheduled
    in the order they are called.

    Arguments:

    input (torch.Tensor required): input like torch.nn.functional.linear

    weight (torch.Tensor required): weight like torch.nn.functional.linear

    bias (torch.Tensor optional): bias like torch.nn.functional.linear

    gradient_accumulation_fusion (bool required): Perform the gradient
        accumulation fusion, requires the custom CUDA extension
        fused_weight_gradient_mlp_cuda module. To use
        gradient_accumulation_fusion you must install APEX with
        --cpp_ext and --cuda_ext. For example: "pip install
        --global-option=\"--cpp_ext\" --global-option=\"--cuda_ext .\"
        " Note that the extension requires CUDA>=11. Otherwise, you
        must turn off gradient accumulation fusion."

    async_grad_allreduce (bool required): Do the allreduce of input
        gradients asyncronously with the computation of weight
        gradients. If sequence_parallel_enabled is True, this must be
        False, as no all reduce is performed.

    sequence_parallel (bool required): Indicates that sequence
        parallelism is used and thus in the forward pass the input is
        all gathered, and the backward pass the input gradients are
        reduce scattered.
    """
    args = [
        input,
        weight,
        bias,
        gradient_accumulation_fusion,
        async_grad_allreduce,
        sequence_parallel,
    ]

    if not linear_with_grad_accumulation_and_async_allreduce.warned:
        if os.environ.get("CUDA_DEVICE_MAX_CONNECTIONS") != "1":
            if sequence_parallel:
                warnings.warn(
                    "When using sequence parallelism it is recommended to set the "
                    "environment variable CUDA_DEVICE_MAX_CONNECTIONS to 1 for "
                    "maximum speedup"
                )
                linear_with_grad_accumulation_and_async_allreduce.warned = True

            if async_grad_allreduce:
                warnings.warn(
                    "When using async grad allreduce it is recommended to set the "
                    "environment variable CUDA_DEVICE_MAX_CONNECTIONS to 1 for "
                    "maximum speedup"
                )
                linear_with_grad_accumulation_and_async_allreduce.warned = True

    return LinearWithGradAccumulationAndAsyncCommunication.apply(*args)


linear_with_grad_accumulation_and_async_allreduce.warned = False


class MergedLinearWithGradAccumulationAndAsyncCommunication(
    torch.autograd.Function
):

    @staticmethod
    @custom_fwd
    def forward(
        ctx,
        input,
        gradient_accumulation_fusion,
        async_grad_allreduce,
        sequence_parallel,
        is_w_parallel,
        *wbs,
    ):
        # disable sequence parallel for now for it requires a global buffer
        assert len(wbs) % 2 == 0
        weights = wbs[::2]
        biases = wbs[1::2]
        assert len(is_w_parallel) == len(weights)
        ctx.save_for_backward(input, *weights)
        ctx.use_bias = tuple(b is not None for b in biases)
        ctx.is_w_parallel = is_w_parallel

        ctx.gradient_accumulation_fusion = gradient_accumulation_fusion
        ctx.async_grad_allreduce = async_grad_allreduce
        ctx.sequence_parallel = sequence_parallel

        if sequence_parallel:
            assert (
                not ctx.async_grad_allreduce
            ), "async_grad_allreduce and sequence_parallel can not be both True"
            world_size = constants.model_parallel_world_size()
            dim_size = list(input.size())
            dim_size[0] = dim_size[0] * world_size

            all_gather_buffer = constants.get_global_memory_buffer().get_tensor(
                dim_size, input.dtype, "mpu"
            )
            torch.distributed._all_gather_base(
                all_gather_buffer, input, group=constants.model_parallel_group()
            )
            total_input = all_gather_buffer
        else:
            total_input = input

        xs = []
        for w, b in zip(weights, biases):
            x = torch.matmul(total_input, w.t())
            if b is not None:
                x = x + b
            xs.append(x)
        return tuple(xs)

    @staticmethod
    @custom_bwd
    def backward(ctx, *grads):
        grads = list(grads)
        input, *weights = ctx.saved_tensors
        assert len(weights) == len(grads)
        use_bias = ctx.use_bias
        is_w_parallel = ctx.is_w_parallel

        if ctx.sequence_parallel:
            world_size = constants.model_parallel_world_size()
            dim_size = list(input.size())
            dim_size[0] = dim_size[0] * world_size

            all_gather_buffer = constants.get_global_memory_buffer().get_tensor(
                dim_size, input.dtype, "mpu"
            )
            handle = torch.distributed._all_gather_base(
                all_gather_buffer,
                input,
                group=constants.model_parallel_group(),
                async_op=True,
            )

            # Here we rely on CUDA_DEVICE_MAX_CONNECTIONS=1 to ensure that the
            # gather is scheduled before the input gradient computation
            total_input = all_gather_buffer
        else:
            total_input = input
        grad_input = 0
        for w, is_parallel, grad in zip(weights, is_w_parallel, grads):
            if is_parallel or constants.model_parallel_rank() == 0:
                grad_input = grad_input + grad.matmul(w)

        if ctx.sequence_parallel:
            handle.wait()

        # Doing gather + slicing during the NeMo forward pass can make this tensor
        # not be contiguous. PyTorch only checks if the tensor is contiguous, and only
        # clones it if it's not contiguous:
        # https://github.com/pytorch/pytorch/blob/c47cf9bc7f9e02f649ab4ed53fe4d35732c92ab6/torch/_refs/__init__.py#L2761
        # Convert the tensor shapes to 2D for execution compatibility
        for i in range(len(grads)):
            grads[i] = grads[i].contiguous().view(-1, grads[i].shape[-1])
        total_input = total_input.view(-1, total_input.shape[-1])

        if ctx.async_grad_allreduce:
            # Asynchronous all-reduce
            handle = torch.distributed.all_reduce(
                grad_input,
                group=constants.model_parallel_group(),
                async_op=True,
            )
            # Here we rely on CUDA_DEVICE_MAX_CONNECTIONS=1 to ensure that the
            # all-reduce is scheduled before the weight gradient computation

        if ctx.sequence_parallel:
            assert not ctx.async_grad_allreduce
            dim_size = list(input.size())
            sub_grad_input = torch.empty(
                dim_size,
                dtype=input.dtype,
                device=torch.cuda.current_device(),
                requires_grad=False,
            )
            # reduce_scatter
            handle = torch.distributed._reduce_scatter_base(
                sub_grad_input,
                grad_input,
                group=constants.model_parallel_group(),
                async_op=True,
            )
            # Here we rely on CUDA_DEVICE_MAX_CONNECTIONS=1 to ensure that the
            # reduce scatter is scheduled before the weight gradient computation

        if ctx.gradient_accumulation_fusion:
            gws = []
            for weight, grad_output in zip(weights, grads):
                if weight.main_grad.dtype == torch.float32:
                    fused_weight_gradient_mlp_cuda.wgrad_gemm_accum_fp32(
                        total_input, grad_output, weight.main_grad
                    )
                elif weight.main_grad.dtype in (torch.float16, torch.bfloat16):
                    fused_weight_gradient_mlp_cuda.wgrad_gemm_accum_fp16(
                        total_input, grad_output, weight.main_grad
                    )
                else:
                    raise RuntimeError(
                        "Unsupported gradient type for gradient accumulation fusion"
                    )
                if hasattr(weight, "grad_added_to_main_grad"):
                    # When overlap_grad_reduce is True, need to ensure that backward hooks
                    # are all run on the main backprop thread to prevent deadlocks. Setup
                    # dummy grad_weight tensor to prevent backward hooks from being run
                    # in a background thread.
                    if getattr(weight, "zero_out_wgrad", False):
                        grad_weight = torch.zeros(
                            weight.main_grad.shape,
                            dtype=input.dtype,
                            device=torch.cuda.current_device(),
                            requires_grad=False,
                        )
                    else:
                        grad_weight = torch.empty(
                            weight.main_grad.shape,
                            dtype=input.dtype,
                            device=torch.cuda.current_device(),
                            requires_grad=False,
                        )
                    weight.grad_added_to_main_grad = True
                else:
                    grad_weight = None
                gws.append(grad_weight)
        else:
            gws = []
            for w, g in zip(weights, grads):
                gws.append(g.t().matmul(total_input))
        gbs = [
            g.sum(dim=0) if use_bias[i] else None for i, g in enumerate(grads)
        ]

        if ctx.sequence_parallel:
            handle.wait()
            return (
                sub_grad_input,
                None,
                None,
                None,
                None,
                *list(itertools.chain.from_iterable(zip(gws, gbs))),
            )

        if ctx.async_grad_allreduce:
            handle.wait()

        return (
            grad_input,
            None,
            None,
            None,
            None,
            *list(itertools.chain.from_iterable(zip(gws, gbs))),
        )


def merged_linear_with_grad_accumulation_and_async_allreduce(
    input: torch.Tensor,
    gradient_accumulation_fusion: bool,
    async_grad_allreduce: bool,
    sequence_parallel: bool,
    is_w_parallel: List[bool],
    *wbs: List[torch.Tensor | None],
) -> torch.Tensor:
    """Similar to linear_with_grad_accumulation_and_async_allreduce
    but does multiple linear-layer forward/backward calls with a single all gather operation.
    """
    args = [
        input,
        gradient_accumulation_fusion,
        async_grad_allreduce,
        sequence_parallel,
        is_w_parallel,
        *wbs,
    ]

    if not merged_linear_with_grad_accumulation_and_async_allreduce.warned:
        if os.environ.get("CUDA_DEVICE_MAX_CONNECTIONS") != "1":
            if sequence_parallel:
                warnings.warn(
                    "When using sequence parallelism it is recommended to set the "
                    "environment variable CUDA_DEVICE_MAX_CONNECTIONS to 1 for "
                    "maximum speedup"
                )
                merged_linear_with_grad_accumulation_and_async_allreduce.warned = (
                    True
                )

            if async_grad_allreduce:
                warnings.warn(
                    "When using async grad allreduce it is recommended to set the "
                    "environment variable CUDA_DEVICE_MAX_CONNECTIONS to 1 for "
                    "maximum speedup"
                )
                merged_linear_with_grad_accumulation_and_async_allreduce.warned = (
                    True
                )

    return MergedLinearWithGradAccumulationAndAsyncCommunication.apply(*args)


merged_linear_with_grad_accumulation_and_async_allreduce.warned = False


class ColumnParallelLinear(torch.nn.Module):
    """Linear layer with column parallelism.

    The linear layer is defined as Y = XA + b. A is parallelized along
    its second dimension as A = [A_1, ..., A_p].

    Arguments:
        input_size: first dimension of matrix A.
        output_size: second dimension of matrix A.

    Keyword Arguments
        bias: If true, add bias
        gather_output: If true, call all-gather on output and make Y available
                       to all GPUs, otherwise, every GPU will have its output
                       which is Y_i = XA_i
        init_method: method to initialize weights. Note that bias is always set
                     to zero.
        stride: For the strided linear layers.
        keep_master_weight_for_test: This was added for testing and should be
                                     set to False. It returns the master weights
                                     used for initialization.
        skip_bias_add: This was added to enable performance optimations where bias
                       can be fused with other elementwise operations. we skip
                       adding bias but instead return it.
        async_tensor_model_parallel_allreduce:
        params_dtype:
        gradient_accumulation_fusion:
    """

    def __init__(
        self,
        input_size,
        output_size,
        bias=True,
        gather_output=False,
        init_method=init.xavier_normal_,
        stride=1,
        skip_bias_add=False,
        async_tensor_model_parallel_allreduce=True,
        perform_initialization=True,
        gradient_accumulation_fusion=False,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super(ColumnParallelLinear, self).__init__()

        # Keep input parameters
        self.input_size = input_size
        self.output_size = output_size
        self.gather_output = gather_output
        # Divide the weight matrix along the last dimension.
        world_size = constants.model_parallel_world_size()
        self.output_size_per_partition = divide(output_size, world_size)
        self.skip_bias_add = skip_bias_add
        assert skip_bias_add is False

        # Parameters.
        # Note: torch.nn.functional.linear performs XA^T + b and as a result
        # we allocate the transpose.
        # Initialize weight.
        # logger.info(
        #     f"ColumnLinear: input_size={input_size}, output_size={output_size}, output_size_per_partition={self.output_size_per_partition}"
        # )
        self.weight = Parameter(
            torch.empty(
                self.output_size_per_partition,
                self.input_size,
                device=device,
                dtype=dtype,
            )
        )
        if perform_initialization:
            _initialize_affine_weight_gpu(
                self.weight, init_method, partition_dim=0, stride=stride
            )

        if bias:
            self.bias = Parameter(
                torch.empty(
                    self.output_size_per_partition, device=device, dtype=dtype
                )
            )
            set_tensor_model_parallel_attributes(self.bias, True, 0, stride)
            # Always initialize bias to zero.
            with torch.no_grad():
                self.bias.zero_()
        else:
            self.register_parameter("bias", None)

        self.async_tensor_model_parallel_allreduce = (
            async_tensor_model_parallel_allreduce and world_size > 1
        )

        if gradient_accumulation_fusion:
            if not _grad_accum_fusion_available:
                raise RuntimeError(
                    "ColumnParallelLinear was called with gradient_accumulation_fusion set "
                    "to True but the custom CUDA extension fused_weight_gradient_mlp_cuda "
                    "module is not found. To use gradient_accumulation_fusion you must "
                    "install APEX with --cpp_ext and --cuda_ext. For example: "
                    'pip install --global-option="--cpp_ext" --global-option="--cuda_ext ." '
                    "Note that the extension requires CUDA>=11. Otherwise, you must turn off "
                    "gradient accumulation fusion."
                )
        self.gradient_accumulation_fusion = gradient_accumulation_fusion

    def forward(self, input_) -> torch.Tensor:
        """Forward of ColumnParallelLinear

        Args:
            input_: 3D tensor whose order of dimension is [sequence, batch, hidden]

        Returns:
            - output
            - bias
        """
        bias = self.bias if not self.skip_bias_add else None
        if (
            self.async_tensor_model_parallel_allreduce
            and constants.sequence_parallel()
        ):
            raise RuntimeError(
                "`async_tensor_model_parallel_allreduce` and `sequence_parallel` "
                "cannot be enabled at the same time."
            )

        if (
            self.async_tensor_model_parallel_allreduce
            or constants.sequence_parallel()
        ):
            input_parallel = input_
        else:
            input_parallel = copy_to_tensor_model_parallel_region(input_)
        # Matrix multiply.
        if not self.weight.requires_grad:
            forward_impl = linear_with_frozen_weight
        else:
            forward_impl = linear_with_grad_accumulation_and_async_allreduce
        output_parallel = forward_impl(
            input=input_parallel,
            weight=self.weight,
            bias=bias,
            gradient_accumulation_fusion=self.gradient_accumulation_fusion,
            async_grad_allreduce=self.async_tensor_model_parallel_allreduce,
            sequence_parallel=constants.sequence_parallel(),
        )
        if self.gather_output:
            # All-gather across the partitions.
            assert not constants.sequence_parallel()
            output = gather_from_tensor_model_parallel_region(output_parallel)
        else:
            output = output_parallel
        # output_bias = self.bias if self.skip_bias_add else None
        return output


class RowParallelLinear(torch.nn.Module):
    """Linear layer with row parallelism.

    The linear layer is defined as Y = XA + b. A is parallelized along
    its first dimension and X along its second dimension as:
               -   -
              | A_1 |
              | .   |
          A = | .   |        X = [X_1, ..., X_p]
              | .   |
              | A_p |
               -   -
    Arguments:
        input_size: first dimension of matrix A.
        output_size: second dimension of matrix A.

    Keyword Arguments:
        bias: If true, add bias. Note that bias is not parallelized.
        input_is_parallel: If true, we assume that the input is already
                           split across the GPUs and we do not split
                           again.
        init_method: method to initialize weights. Note that bias is always set
                     to zero.
        stride: For the strided linear layers.
        keep_master_weight_for_test: This was added for testing and should be
                                     set to False. It returns the master weights
                                     used for initialization.
        skip_bias_add: This was added to enable performance optimization where bias
                       can be fused with other elementwise operations. We skip
                       adding bias but instead return it.
        params_dtype:
        use_cpu_initialization:
        perform_initialization:
        gradient_accumulation_fusion:
        sequence_parallel_enabled:
    """

    def __init__(
        self,
        input_size,
        output_size,
        bias=True,
        input_is_parallel=True,
        init_method=init.xavier_normal_,
        stride=1,
        skip_bias_add=False,
        perform_initialization=True,
        gradient_accumulation_fusion=False,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super(RowParallelLinear, self).__init__()

        # Keep input parameters
        self.input_size = input_size
        self.output_size = output_size
        self.input_is_parallel = input_is_parallel
        # Divide the weight matrix along the last dimension.
        world_size = constants.model_parallel_world_size()
        self.input_size_per_partition = divide(input_size, world_size)
        self.skip_bias_add = skip_bias_add
        self.gradient_accumulation_fusion = gradient_accumulation_fusion
        if constants.sequence_parallel() and not self.input_is_parallel:
            raise RuntimeError(
                "To enable `sequence_parallel`, `input_is_parallel` must be `True`"
            )

        # Parameters.
        # Note: torch.nn.functional.linear performs XA^T + b and as a result
        # we allocate the transpose.
        # Initialize weight.
        self.weight = Parameter(
            torch.empty(
                self.output_size,
                self.input_size_per_partition,
                device=device,
                dtype=dtype,
            )
        )
        if perform_initialization:
            _initialize_affine_weight_gpu(
                self.weight, init_method, partition_dim=1, stride=stride
            )
        if bias:
            self.bias = Parameter(
                torch.empty(self.output_size, device=device, dtype=dtype)
            )
            setattr(self.bias, "sequence_parallel", False)

            # Always initialize bias to zero.
            with torch.no_grad():
                self.bias.zero_()
        else:
            self.register_parameter("bias", None)

    def forward(self, input_) -> torch.Tensor:
        """Forward of RowParallelLinear

        Args:
            input_: 3D tensor whose order of dimension is [sequence, batch, hidden]

        Returns:
            - output
            - bias
        """
        # Set up backprop all-reduce.
        if self.input_is_parallel:
            input_parallel = input_
        else:
            assert not constants.sequence_parallel()
            input_parallel = scatter_to_tensor_model_parallel_region(input_)
        # Matrix multiply.
        if not self.weight.requires_grad:
            _forward_impl = linear_with_frozen_weight
        else:
            _forward_impl = linear_with_grad_accumulation_and_async_allreduce
        output_parallel = _forward_impl(
            input=input_parallel,
            weight=self.weight,
            bias=None,
            gradient_accumulation_fusion=self.gradient_accumulation_fusion,
            async_grad_allreduce=False,
            sequence_parallel=False,  # Here false because we do not need allreduce grad in backward here
        )

        # All-reduce across all the partitions.
        if constants.sequence_parallel():
            output_ = reduce_scatter_to_sequence_parallel_region(
                output_parallel
            )
        else:
            output_ = reduce_from_tensor_model_parallel_region(output_parallel)
        output = output_ + self.bias if self.bias is not None else output_
        return output


def parallel_lm_logits(
    input_: torch.HalfTensor,
    word_embeddings_weight: torch.HalfTensor,
    parallel_output: bool = False,
    async_tensor_model_parallel_allreduce: bool = False,
    gradient_accumulation_fusion: bool = False,
    bias=None,
):
    """LM logits using word embedding weights."""
    sequence_parallel = constants.sequence_parallel()
    # Parallel logits.
    if async_tensor_model_parallel_allreduce or sequence_parallel:
        input_parallel = input_
        model_parallel = constants.model_parallel_world_size() > 1
        async_grad_allreduce = (
            async_tensor_model_parallel_allreduce
            and model_parallel
            and not sequence_parallel
        )
    else:
        input_parallel = copy_to_tensor_model_parallel_region(input_)
        async_grad_allreduce = False

    # Matrix multiply.
    logits_parallel = linear_with_grad_accumulation_and_async_allreduce(
        input=input_parallel,
        weight=word_embeddings_weight,
        bias=bias,
        gradient_accumulation_fusion=gradient_accumulation_fusion,
        async_grad_allreduce=async_grad_allreduce,
        sequence_parallel=sequence_parallel,
    )
    # Gather if needed.

    if parallel_output:
        return logits_parallel

    return gather_from_tensor_model_parallel_region(logits_parallel)


class _VocabParallelCrossEntropy(torch.autograd.Function):

    @staticmethod
    def forward(ctx, vocab_parallel_logits, target, label_smoothing=0.0):
        # Maximum value along vocab dimension across all GPUs.
        logits_max = torch.max(vocab_parallel_logits, dim=-1)[0]
        torch.distributed.all_reduce(
            logits_max,
            op=torch.distributed.ReduceOp.MAX,
            group=constants.model_parallel_group(),
        )
        # Subtract the maximum value.
        vocab_parallel_logits = vocab_parallel_logits - logits_max.unsqueeze(
            dim=-1
        )

        # Get the partition's vocab indecies
        get_vocab_range = VocabUtility.vocab_range_from_per_partition_vocab_size
        partition_vocab_size = vocab_parallel_logits.size()[-1]
        rank = constants.model_parallel_rank()
        world_size = constants.model_parallel_world_size()
        vocab_start_index, vocab_end_index = get_vocab_range(
            partition_vocab_size, rank, world_size
        )

        # Create a mask of valid vocab ids (1 means it needs to be masked).
        target_mask = (target < vocab_start_index) | (target >= vocab_end_index)
        masked_target = target.clone() - vocab_start_index
        masked_target[target_mask] = 0

        # Get predicted-logits = logits[target].
        # For Simplicity, we convert logits to a 2-D tensor with size
        # [*, partition-vocab-size] and target to a 1-D tensor of size [*].
        logits_2d = vocab_parallel_logits.view(-1, partition_vocab_size)
        masked_target_1d = masked_target.view(-1)
        arange_1d = torch.arange(
            start=0, end=logits_2d.size()[0], device=logits_2d.device
        )
        predicted_logits_1d = logits_2d[arange_1d, masked_target_1d]
        predicted_logits_1d = predicted_logits_1d.clone().contiguous()
        predicted_logits = predicted_logits_1d.view_as(target)
        predicted_logits[target_mask] = 0.0
        # All reduce is needed to get the chunks from other GPUs.
        torch.distributed.all_reduce(
            predicted_logits,
            op=torch.distributed.ReduceOp.SUM,
            group=constants.model_parallel_group(),
        )

        # Sum of exponential of logits along vocab dimension across all GPUs.
        exp_logits = vocab_parallel_logits
        torch.exp(vocab_parallel_logits, out=exp_logits)
        sum_exp_logits = exp_logits.sum(dim=-1)
        torch.distributed.all_reduce(
            sum_exp_logits,
            op=torch.distributed.ReduceOp.SUM,
            group=constants.model_parallel_group(),
        )

        # Loss = log(sum(exp(logits))) - predicted-logit.
        loss = torch.log(sum_exp_logits) - predicted_logits

        # Normalize and optionally smooth logits
        exp_logits.div_(sum_exp_logits.unsqueeze(dim=-1))

        vocab_size = exp_logits.size(-1)
        if label_smoothing > 0:
            """
            We'd like to assign 1 / (K - 1) probability mass to every index that is not the ground truth.
            = (1 - alpha) * y_gt + alpha * mean(y_{i for i != gt})
            = (1 - alpha) * y_gt + (alpha / (K - 1)) * \sum_{i != gt} y_i
            = ((K - 1) * (1 - alpha) / (K - 1)) * y_gt + (alpha / (K - 1)) * \sum_{i != gt} y_i
            = (K * (1 - alpha) - 1) / (K - 1)) * y_gt  + (alpha / (K - 1)) * \sum_{i} y_i
            = (1 - (alpha * K) / (K - 1)) * y_gt + ( (alpha * K) / (K - 1) ) * \sum_{i} y_i / K
            From: https://github.com/NVIDIA/NeMo/blob/main/nemo/collections/common/losses/smoothed_cross_entropy.py
            """
            assert 1.0 > label_smoothing > 0.0
            smoothing = label_smoothing * vocab_size / (vocab_size - 1)

            # Exp logits at this point are normalized probabilities. So we can just take the log to get log-probs.
            log_probs = torch.log(exp_logits)
            mean_log_probs = log_probs.mean(dim=-1)
            loss = (1.0 - smoothing) * loss - smoothing * mean_log_probs

        ctx.label_smoothing, ctx.vocab_size = label_smoothing, vocab_size

        # Store softmax, target-mask and masked-target for backward pass.
        ctx.save_for_backward(exp_logits, target_mask, masked_target_1d)

        return loss

    @staticmethod
    def backward(ctx, grad_output):
        # Retreive tensors from the forward path.
        softmax, target_mask, masked_target_1d = ctx.saved_tensors
        label_smoothing, vocab_size = ctx.label_smoothing, ctx.vocab_size

        # All the inputs have softmax as thier gradient.
        grad_input = softmax
        # For simplicity, work with the 2D gradient.
        partition_vocab_size = softmax.size()[-1]
        grad_2d = grad_input.view(-1, partition_vocab_size)

        # Add the gradient from matching classes.
        arange_1d = torch.arange(
            start=0, end=grad_2d.size()[0], device=grad_2d.device
        )

        softmax_update = 1.0 - target_mask.view(-1).float()

        if label_smoothing > 0:
            smoothing = label_smoothing * vocab_size / (vocab_size - 1)
            grad_2d[arange_1d, masked_target_1d] -= (
                1.0 - smoothing
            ) * softmax_update
            average_grad = 1 / vocab_size
            grad_2d[arange_1d, :] -= smoothing * average_grad
        else:
            grad_2d[arange_1d, masked_target_1d] -= softmax_update

        # Finally elementwise multiplication with the output gradients.
        grad_input.mul_(grad_output.unsqueeze(dim=-1))

        return grad_input, None, None


def vocab_parallel_cross_entropy(
    vocab_parallel_logits, target, label_smoothing=0.0
):
    """
    Performs cross entropy loss when logits are split across tensor parallel ranks

    Arguments:
        vocab_parallel_logits: logits split across tensor parallel ranks
                               dimension is [sequence_length, batch_size, hidden_size]

        target: correct vocab ids of dimseion [sequence_length, micro_batch_size]

        lobal_smoothing: smoothing factor, must be in range [0.0, 1.0)
                         default is no smoothing (=0.0)
    """
    return _VocabParallelCrossEntropy.apply(
        vocab_parallel_logits, target, label_smoothing
    )



from contextlib import contextmanager
from typing import *
import dataclasses

try:
    from megatron.core import parallel_state
    from megatron.core.distributed.distributed_data_parallel import (
        DistributedDataParallel,
    )
    from megatron.core.distributed.finalize_model_grads import (
        finalize_model_grads,
    )
    from megatron.core.distributed.param_and_grad_buffer import (
        ParamAndGradBuffer,
    )
    from megatron.core.optimizer import get_megatron_optimizer
    from megatron.core.optimizer.optimizer_config import OptimizerConfig
    from megatron.core.transformer.transformer_config import TransformerConfig
except ImportError or ModuleNotFoundError:
    # dummy class for type hint, due to missing files in megatron CPU installation
    class TransformerConfig:
        pass


import torch
import torch.distributed as dist
import torch.nn as nn
import transformers

from realhf.api.core import model_api
from realhf.base import constants, logging
from realhf.impl.model.backend.pipe_runner import PipelineRunner
from realhf.impl.model.backend.utils import (
    MegatronEngine,
    OptimizerParamScheduler,
)
from realhf.impl.model.modules.mlp import get_activation_fn
from realhf.impl.model.nn.real_llm_api import ReaLModel
from realhf.impl.model.nn.real_llm_generate import GenerationConfig

WITHIN_MEGATRON_CONTEXT = False

logger = logging.getLogger("Megatron Backend", "benchmark")


@contextmanager
def megatron_ctx():
    global WITHIN_MEGATRON_CONTEXT
    if WITHIN_MEGATRON_CONTEXT:
        raise RuntimeError(
            "Megatron context is already set up. Destroy it first."
        )

    WITHIN_MEGATRON_CONTEXT = True

    grid = constants.grid()
    # TODO: implement context parallel.
    # TODO: implement expert parallel.

    # Build the data-parallel groups.
    g = constants.data_parallel_group()
    parallel_state._DATA_PARALLEL_GROUP = g
    parallel_state._DATA_PARALLEL_GROUP_GLOO = (
        grid.get_data_parallel_group_gloo()
    )
    parallel_state._DATA_PARALLEL_GLOBAL_RANKS = dist.get_process_group_ranks(g)
    parallel_state._DATA_PARALLEL_GROUP_WITH_CP = g
    parallel_state._DATA_PARALLEL_GROUP_WITH_CP_GLOO = (
        grid.get_data_parallel_group_gloo()
    )
    parallel_state._DATA_PARALLEL_GLOBAL_RANKS_WITH_CP = (
        dist.get_process_group_ranks(g)
    )

    # Build the context-parallel groups.
    parallel_state._CONTEXT_PARALLEL_GROUP = constants.self_group()
    parallel_state._CONTEXT_PARALLEL_GLOBAL_RANKS = [dist.get_rank()]

    # Build the model-parallel groups.
    parallel_state._MODEL_PARALLEL_GROUP = grid.get_model_parallel_group()

    # Build the tensor model-parallel groups.
    g = constants.model_parallel_group()
    parallel_state._TENSOR_MODEL_PARALLEL_GROUP = g

    # Build the pipeline model-parallel groups and embedding groups
    # (first and last rank in each pipeline model-parallel group).
    g = constants.pipe_parallel_group()
    parallel_state._PIPELINE_MODEL_PARALLEL_GROUP = g
    parallel_state._PIPELINE_GLOBAL_RANKS = dist.get_process_group_ranks(g)
    parallel_state._EMBEDDING_GROUP = grid.embedding_proc_group
    parallel_state._EMBEDDING_GLOBAL_RANKS = (
        dist.get_process_group_ranks(grid.embedding_proc_group)
        if grid.embedding_proc_group is not None
        else list(range(dist.get_world_size()))
    )
    parallel_state._POSITION_EMBEDDING_GROUP = (
        grid.position_embedding_proc_group
    )
    parallel_state._POSITION_EMBEDDING_GLOBAL_RANKS = (
        dist.get_process_group_ranks(grid.position_embedding_proc_group)
        if grid.position_embedding_proc_group is not None
        else list(range(dist.get_world_size()))
    )

    # Build the tensor + data parallel groups.
    parallel_state._TENSOR_AND_DATA_PARALLEL_GROUP = grid.tp_dp_proc_group
    parallel_state._TENSOR_AND_DATA_PARALLEL_GROUP_WITH_CP = (
        grid.tp_dp_proc_group
    )

    # Build the tensor + expert parallel groups
    parallel_state._EXPERT_MODEL_PARALLEL_GROUP = constants.self_group()
    parallel_state._TENSOR_AND_EXPERT_PARALLEL_GROUP = (
        constants.model_parallel_group()
    )
    g = constants.data_parallel_group()
    parallel_state._DATA_MODULO_EXPERT_PARALLEL_GROUP = g
    parallel_state._DATA_MODULO_EXPERT_PARALLEL_GROUP_GLOO = (
        grid.get_data_parallel_group_gloo()
    )

    # Remove the global memory buffer for megatron to save GPU memory.
    parallel_state._GLOBAL_MEMORY_BUFFER = None
    yield
    WITHIN_MEGATRON_CONTEXT = False


def get_megatron_transformer_config(
    mconfig: model_api.ReaLModelConfig,
) -> TransformerConfig:
    nq = mconfig.hidden_dim // mconfig.head_dim
    n_group = nq // mconfig.n_kv_heads
    return TransformerConfig(
        num_layers=mconfig.n_layers,
        hidden_size=mconfig.hidden_dim,
        num_attention_heads=nq,
        num_query_groups=n_group,
        ffn_hidden_size=mconfig.intermediate_dim,
        kv_channels=mconfig.n_kv_heads,
        hidden_dropout=0.0,
        attention_dropout=mconfig.attn_pdrop,
        layernorm_epsilon=mconfig.layer_norm_epsilon,
        add_qkv_bias=mconfig.use_attention_bias,
        activation_func=get_activation_fn(mconfig.activation_function),
        rotary_interleaved=mconfig.rotary_interleaved,
        normalization=(
            "RMSNorm" if mconfig.layer_norm_type == "rms" else "LayerNorm"
        ),
        attention_softmax_in_fp32=True,
        apply_query_key_layer_scaling=mconfig.scale_attn_by_inverse_layer_idx,
    )


class ReaLMegatronEngine:
    # TODO: merge this with deepspeed and inference engine

    def __init__(self, module: ReaLModel, megatron_engine: MegatronEngine):
        self.module = module

        self.device = module.device
        self.dtype = module.dtype

        self.engine = megatron_engine

        if constants.pipe_parallel_world_size() > 1:
            self.pipe_runner = PipelineRunner(module)

            model_parameters = filter(
                lambda p: p.requires_grad, self.module.parameters()
            )
            num_params = sum([p.numel() for p in model_parameters])
            unique_params = num_params

            params_tensor = torch.LongTensor(
                data=[num_params, unique_params]
            ).to(self.device)
            dist.all_reduce(
                params_tensor, group=constants.grid().get_model_parallel_group()
            )
            params_tensor = params_tensor.tolist()
            total_params = params_tensor[0]
            unique_params = params_tensor[1]

            if constants.parallelism_rank() == 0:
                logger.info(
                    f"CONFIG: default_train_mbs={self.pipe_runner.default_train_mbs} "
                    f"default_inf_mbs={self.pipe_runner.default_inf_mbs} "
                    f"num_layers(this stage)={self.module.num_layers} "
                    f"pp_size={constants.pipe_parallel_world_size()} "
                    f"dp_size={constants.data_parallel_world_size()} "
                    f"mp_size={constants.model_parallel_world_size()} "
                )
            if constants.data_parallel_rank() == 0:
                logger.info(
                    f"rank={constants.parallelism_rank()} "
                    f"stage={constants.pipe_parallel_rank()} "
                    f"layers={self.module.num_layers} "
                    f"[{self.module.layer_idx_start}, {self.module.layer_idx_end}) "
                    f"stage_params={num_params} ({num_params/1e6:0.3f}M) "
                    f"total_params={total_params} ({total_params/1e6:0.3f}M) "
                    f"unique_params={unique_params} ({unique_params/1e6:0.3f}M)"
                )

    def train(self, mode: bool = True):
        self.module.train(mode)
        self.engine.ddp.train(mode)
        return self

    def eval(self):
        self.module.eval()
        self.engine.ddp.eval()
        return self

    def train_batch(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        loss_fn: Callable,
        version_steps: int,
        input_lens_for_partition: Optional[torch.Tensor] = None,
        num_micro_batches: Optional[int] = None,
        **loss_fn_kwargs,
    ):
        with megatron_ctx():
            self.engine.zero_grad()
            if constants.pipe_parallel_world_size() > 1:
                return self.pipe_runner.train_batch(
                    engine=self.engine,
                    seqlens_cpu=seqlens_cpu,
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                    loss_fn=loss_fn,
                    version_steps=version_steps,
                    input_lens_for_partition=input_lens_for_partition,
                    num_micro_batches=num_micro_batches,
                    **loss_fn_kwargs,
                )
            else:
                # FIXME: num_micro_batches is ignored here
                max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))
                model_output = self.engine.ddp(
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                    max_seqlen=max_seqlen,
                ).logits
                loss, stat = loss_fn(
                    model_output, packed_input_ids, cu_seqlens, **loss_fn_kwargs
                )
                self.engine.optim.scale_loss(loss).backward()
                finalize_model_grads([self.engine.ddp])
                update_successful, grad_norm, _ = self.engine.optim.step()
                if update_successful:
                    self.engine.lr_scheduler.step_absolute(version_steps)
                if (
                    constants.data_parallel_rank() == 0
                    and constants.model_parallel_rank() == 0
                ):
                    logger.info(
                        f"Megatron backend update success? {update_successful}. "
                        f"Grad Norm: {grad_norm}. "
                        f"Current loss scale: {self.engine.optim.get_loss_scale()}. "
                    )
                return stat

    @torch.no_grad()
    def eval_batch(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        loss_fn: Callable,
        input_lens_for_partition: Optional[torch.Tensor] = None,
        num_micro_batches: Optional[int] = None,
        **loss_fn_kwargs,
    ):
        with megatron_ctx():
            if constants.pipe_parallel_world_size() > 1:
                return self.pipe_runner.eval_batch(
                    seqlens_cpu=seqlens_cpu,
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                    loss_fn=loss_fn,
                    input_lens_for_partition=input_lens_for_partition,
                    num_micro_batches=num_micro_batches,
                    **loss_fn_kwargs,
                )
            else:
                max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))
                model_output = self.engine.ddp(
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                    max_seqlen=max_seqlen,
                ).logits
                _, stat = loss_fn(
                    model_output, packed_input_ids, cu_seqlens, **loss_fn_kwargs
                )
                return stat

    def forward(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        num_micro_batches: Optional[int] = None,
    ):
        with megatron_ctx():
            if constants.pipe_parallel_world_size() > 1:
                return self.pipe_runner.forward(
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                    seqlens_cpu=seqlens_cpu,
                    num_micro_batches=num_micro_batches,
                )
            else:
                max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))
                return self.engine.ddp(
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                    max_seqlen=max_seqlen,
                ).logits

    @torch.no_grad()
    def generate(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        tokenizer: transformers.PreTrainedTokenizerFast,
        gconfig: GenerationConfig = dataclasses.field(
            default_factory=GenerationConfig
        ),
        num_micro_batches: Optional[int] = None,
    ):
        with megatron_ctx():
            if constants.pipe_parallel_world_size() > 1:
                return self.pipe_runner.generate(
                    seqlens_cpu=seqlens_cpu,
                    num_micro_batches=num_micro_batches,
                    tokenizer=tokenizer,
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                    gconfig=gconfig,
                )
            else:
                max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))
                res = self.module.generate(
                    tokenizer=tokenizer,
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                    max_seqlen=max_seqlen,
                    gconfig=gconfig,
                )
                return res.sequences, res.scores, res.logits_mask


@dataclasses.dataclass
class MegatronTrainBackend(model_api.ModelBackend):
    optimizer_name: str = dataclasses.field(
        metadata={"choices": ["adam", "sgd"]},
        default="adam",
    )
    optimizer_config: dict = dataclasses.field(
        default_factory=lambda: dict(
            lr=1e-5, weight_decay=0.1, betas=(0.9, 0.95), eps=1e-5
        )
    )
    lr_scheduler_type: str = "cosine"
    warmup_steps_proportion: float = 0.0
    min_lr_ratio: float = 0.0  # will be used for linear and cosine schedule
    enable_fp16: bool = True
    enable_bf16: bool = False
    use_zero_optimization: bool = True
    overlap_grad_reduce: bool = True
    overlap_param_gather: bool = True
    accumulate_allreduce_grads_in_fp32: bool = False
    initial_loss_scale: float = 4096.0
    gradient_clipping: float = 1.0
    # addtional args
    additional_config: Dict = dataclasses.field(default_factory=dict)

    def _initialize(
        self, model: model_api.Model, spec: model_api.FinetuneSpec
    ) -> model_api.Model:
        module = model.module
        if not isinstance(module, ReaLModel):
            raise ValueError("MegatronTrainBackend only supports ReaLModel.")
        with megatron_ctx():
            module = DistributedDataParallel(
                config=get_megatron_transformer_config(module.config),
                module=module,
                data_parallel_group=constants.data_parallel_group(),
                accumulate_allreduce_grads_in_fp32=self.accumulate_allreduce_grads_in_fp32,
                overlap_grad_reduce=self.overlap_grad_reduce,
                use_distributed_optimizer=self.use_zero_optimization,
                expert_data_parallel_group=None,
                disable_bucketing=False,
                check_for_nan_in_grad=False,
            )

        real_model: ReaLModel = module.module
        if self.use_zero_optimization:
            # Remap parameters.
            assert len(module.buffers) == 1
            param_grad_buf: ParamAndGradBuffer = module.buffers[0]

            # Map Megatron flattened parameters to ReaLModel!
            real_model.contiguous_param = param_grad_buf.param_data

            # Sanity checks.
            assert real_model._param_size == param_grad_buf.numel
            for n, p in real_model.layers.named_parameters():
                n = ".".join(
                    [
                        str(real_model.layer_idx_start + int(n.split(".")[0])),
                        n.split(".", 1)[1],
                    ]
                )
                idx_start, idx_end, _ = param_grad_buf.param_index_map[p]
                assert real_model._param_spec[n].start_idx == idx_start
                assert real_model._param_spec[n].end_idx == idx_end
                assert real_model._param_spec[n].shape == p.shape
                assert torch.allclose(
                    p,
                    real_model.contiguous_param[idx_start:idx_end].view(
                        p.shape
                    ),
                )

        betas = self.optimizer_config.get("betas", (0.9, 0.95))
        wd = self.optimizer_config.get("weight_decay", 0.1)
        lr = self.optimizer_config["lr"]
        opt_cfg = OptimizerConfig(
            optimizer=self.optimizer_name,
            fp16=self.enable_fp16,
            bf16=self.enable_bf16,
            lr=lr,
            min_lr=self.min_lr_ratio * lr,
            weight_decay=wd,
            params_dtype=real_model.dtype,
            initial_loss_scale=self.initial_loss_scale,
            adam_beta1=betas[0],
            adam_beta2=betas[1],
            adam_eps=self.optimizer_config.get("eps", 1e-5),
            sgd_momentum=self.optimizer_config.get("momentum", 0.9),
            use_distributed_optimizer=self.use_zero_optimization,
            overlap_grad_reduce=self.overlap_grad_reduce,
            overlap_param_gather=self.overlap_param_gather,
            clip_grad=self.gradient_clipping,
        )

        with megatron_ctx():
            # no_weight_decay_cond and scale_lr_cond have the following signature:
            # foo(name: str, param: torch.Tensor) -> bool
            optimizer = get_megatron_optimizer(
                opt_cfg,
                [module],
                no_weight_decay_cond=lambda n, p: any(
                    k in n for k in ["bias", "ln.weight", "ln_f.weight"]
                ),
                scale_lr_cond=None,
                lr_mult=1.0,
            )

            warmup_steps = int(
                self.warmup_steps_proportion * spec.total_train_steps
            )
            lr_scheduler = OptimizerParamScheduler(
                optimizer,
                init_lr=0.0 if self.warmup_steps_proportion > 0 else lr,
                max_lr=lr,
                min_lr=self.min_lr_ratio * lr,
                lr_warmup_steps=warmup_steps,
                lr_decay_steps=spec.total_train_steps - warmup_steps,
                lr_decay_style=self.lr_scheduler_type,
                start_wd=wd,
                end_wd=wd,
                wd_incr_steps=spec.total_train_steps,
                wd_incr_style="constant",
            )

        mg_engine = MegatronEngine(module, optimizer, lr_scheduler)
        model.module = ReaLMegatronEngine(real_model, mg_engine)
        return model


model_api.register_backend("megatron", MegatronTrainBackend)



from typing import Optional, Tuple

import numpy as np
import torch
import torch.distributed as dist
import transformers

from realhf.base import constants, logging
from realhf.impl.model.utils.padding import pad_input, unpad_input

logger = logging.getLogger("Modeling Functional Utils")


@torch.jit.script
def upcast_masked_softmax(
    x: torch.Tensor,
    mask: torch.Tensor,
    mask_value: torch.Tensor,
    scale: float,
    softmax_dtype: torch.dtype,
):
    input_dtype = x.dtype
    x = x.to(softmax_dtype) * scale
    x = torch.where(mask, x, mask_value)
    x = torch.nn.functional.softmax(x, dim=-1).to(input_dtype)
    return x


@torch.jit.script
def upcast_softmax(x: torch.Tensor, scale: float, softmax_dtype: torch.dtype):
    input_dtype = x.dtype
    x = x.to(softmax_dtype) * scale
    x = torch.nn.functional.softmax(x, dim=-1).to(input_dtype)
    return x


@torch.jit.script
def masked_softmax(
    x: torch.Tensor, mask: torch.Tensor, mask_value: torch.Tensor
):
    x = torch.where(mask, x, mask_value)
    x = torch.nn.functional.softmax(x, dim=-1)
    return x


def repeat_kv(x: torch.Tensor, n_rep: int) -> torch.Tensor:
    """torch.repeat_interleave(x, dim=2, repeats=n_rep)"""
    total_seqlen, n_kv_heads, head_dim = x.shape
    if n_rep == 1:
        return x
    return (
        x[:, :, None, :]
        .expand(total_seqlen, n_kv_heads, n_rep, head_dim)
        .reshape(total_seqlen, n_kv_heads * n_rep, head_dim)
    )


def mask_eos_token(
    logits: torch.Tensor,
    eos_token_id: Optional[int] = None,
) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    # for min_new_tokens
    if eos_token_id is not None:
        logits[..., eos_token_id] = torch.finfo(logits.dtype).min
    return logits


def gather_shifted_log_probs(
    logits: torch.FloatTensor, labels: torch.LongTensor
) -> torch.FloatTensor:
    """Gather log probs of shifted labels from logits.

    Args:
        logits (torch.FloatTensor): Non-shifted logits with shape [bs, seqlen].
            The final value at [:, seqlen -1] is not used.
        labels (torch.LongTensor): Non-shifted labels/input_ids with shape [bs, seqlen].
            The first value at [:, 0] has no corresponding log prob.

    Returns:
        torch.FloatTensor: Shifted log probability with shape [bs, seqlen -1].
    """
    logits = logits[:, :-1]
    labels = labels[:, 1:]
    log_probs = torch.nn.functional.log_softmax(logits, dim=-1)
    log_probs_labels = log_probs.gather(dim=-1, index=labels.unsqueeze(-1))
    return log_probs_labels.squeeze(-1)


def build_shift_one_indices(
    x: torch.HalfTensor, cu_seqlens: torch.IntTensor
) -> torch.IntTensor:
    """Build indices for shifting labels/input_ids one step to the left.

    Equivalent to:
    ```
    shift_one_indices = torch.cat([
        torch.arange(cu_seqlens[i] + 1, cu_seqlens[i + 1], dtype=torch.long, device=cu_seqlens.device)
        for i in range(cu_seqlens.shape[0] - 1)
    ])
    ```
    but the above implementaion will implicitly convert a tensor (cu_seqlens[i]) to an integer,
    which will cause a cuda device sync and slow down performance.

    Args:
        x (torch.HalfTensor): Shape [total_seqlen]. This tensor is required to get
            total_seqlen from its shape. Computing total_seqlen from cu_seqlens will implicitly cause
            a cuda device sync.
        cu_seqlens (torch.IntTensor): Shape [bs + 1]. Indices marking the start
            and end of each sequences.

    Returns:
        torch.IntTensor: Shape [tot_seqlen - bs]. Indices for shifting labels/input_ids
            one step to the left.
    """
    total_seqlen = x.shape[0]
    bs = cu_seqlens.shape[0] - 1
    short1lens = cu_seqlens[1:] - cu_seqlens[:-1] - 1
    short1cu_seqlens = torch.nn.functional.pad(
        short1lens.cumsum(0), (1, 0), value=0
    )
    indexing_t = torch.arange(
        total_seqlen - bs, dtype=torch.long, device=cu_seqlens.device
    )
    return indexing_t + (
        indexing_t.unsqueeze(0) >= short1cu_seqlens[:-1].unsqueeze(1)
    ).sum(0)


def build_leave_one_indices(
    x: torch.HalfTensor, cu_seqlens: torch.IntTensor
) -> torch.IntTensor:
    """Build indices for leaving one token out at the end of each sequence.

    Equivalent to:
    ```
    leave_one_indices = torch.cat([
        torch.arange(cu_seqlens[i], cu_seqlens[i + 1] - 1, dtype=torch.long, device=cu_seqlens.device)
        for i in range(cu_seqlens.shape[0] - 1)
    ])
    ```
    but the above implementaion will implicitly convert a tensor (cu_seqlens[i]) to an integer,
    which will cause a cuda device sync and slow down performance.

    Args:
        x (torch.HalfTensor): Shape [total_seqlen]. This tensor is required to get
            total_seqlen from its shape. Computing total_seqlen from cu_seqlens will implicitly cause
            a cuda device sync.
        cu_seqlens (torch.IntTensor): Shape [bs + 1]. Indices marking the start
            and end of each sequences.

    Returns:
        torch.IntTensor: Shape [tot_seqlen - bs]. Indices for shifting labels/input_ids
            one step to the left.
    """
    total_seqlen = x.shape[0]
    bs = cu_seqlens.shape[0] - 1
    short1lens = cu_seqlens[1:] - cu_seqlens[:-1] - 1
    short1cu_seqlens = torch.nn.functional.pad(
        short1lens.cumsum(0), (1, 0), value=0
    )
    indexing_t = torch.arange(
        total_seqlen - bs, dtype=torch.long, device=cu_seqlens.device
    )
    return (
        indexing_t
        + (indexing_t.unsqueeze(0) >= short1cu_seqlens[:-1].unsqueeze(1)).sum(0)
        - 1
    )


def gather_packed_shifted_log_probs(
    logits: torch.FloatTensor,
    cu_seqlens: torch.Tensor,
    labels: torch.LongTensor,
) -> torch.FloatTensor:
    """Gather log probs from packed input_ids and logits.

    Args:
        logits_ (torch.FloatTensor): Shape [tot_seqlen]. The final value at the end of
            each sequence is not used.
        cu_seqlens (torch.Tensor): Shape [#seqs + 1]. Indices marking the start
            and end of each sequences.
        labels_ (torch.LongTensor): Labels or input_ids with shape [tot_seqlen].
            The first value at the beginning of each sequence has no corresponding log prob.

    Returns:
        torch.FloatTensor: Log probability with shape [tot_seqlen - #seqs].
    """
    labels = torch.nn.functional.pad(labels[1:], (0, 1), value=0)
    leave_one_indices = build_leave_one_indices(logits, cu_seqlens)
    if constants.model_parallel_world_size() > 1:
        # NOTE: logprobs is freaking sensitive to input_ids. If the input sequence is a natural sequence, everything will be fine.
        # However, if we input random token IDs, parallel cross entropy can produce VERY different results than the normal
        # torch.gather based version (e.g., the maximum absolute different can reach ~50).
        from realhf.impl.model.parallelism.model_parallel.modules import (
            vocab_parallel_cross_entropy,
        )

        logprobs = -vocab_parallel_cross_entropy(logits, labels)[
            leave_one_indices
        ]
        ########### sanity check ###########
        # world_size = constants.model_parallel_world_size()
        # dim_size = [logits.shape[1] * world_size, logits.shape[0]]
        # all_gather_buffer = torch.zeros(*dim_size, dtype=logits.dtype, device=logits.device)
        # torch.distributed._all_gather_base(
        #     all_gather_buffer,
        #     logits.transpose(0, 1).contiguous(),
        #     group=constants.model_parallel_group(),
        # )
        # logits2 = all_gather_buffer.transpose(0, 1).contiguous()
        # logprobs2 = gather_packed_shifted_log_probs(logits2, cu_seqlens, packed_input_ids).float()
        # assert torch.allclose(logprobs, logprobs2, atol=2e-2), (
        #     (logprobs - logprobs2).abs().max(),
        #     logprobs,
        #     logprobs2,
        # )
        ########### sanity check ###########
        return logprobs
    logits_shape = logits.shape
    # shift_one_indices = torch.cat([
    #     torch.arange(cu_seqlens[i] + 1 , cu_seqlens[i + 1], dtype=torch.long, device=cu_seqlens.device)
    #     for i in range(cu_seqlens.shape[0] - 1)
    # ])
    # shift labels one step to the left and pad it to match the shape of logits
    log_probs = torch.nn.functional.log_softmax(logits, dim=-1)
    log_probs_labels = log_probs.gather(
        dim=-1, index=labels.unsqueeze(-1)
    ).squeeze(-1)
    log_probs_labels = log_probs_labels[leave_one_indices]
    assert (
        log_probs_labels.shape[0] == logits_shape[0] - cu_seqlens.shape[0] + 1
    ), (
        log_probs_labels.shape,
        logits_shape,
        cu_seqlens.shape,
        cu_seqlens,
        # shift_one_indices,
    )
    return log_probs_labels


def apply_logits_mask(logits: torch.HalfTensor, mask: torch.BoolTensor):
    assert (
        mask.shape[-1]
        == logits.shape[-1] * constants.model_parallel_world_size()
    ), (
        constants.model_parallel_world_size(),
        logits.shape,
        mask.shape,
    )
    parallel_vocab_size = logits.shape[-1]
    mp_rank = constants.model_parallel_rank()
    mask = mask[
        :, mp_rank * parallel_vocab_size : (mp_rank + 1) * parallel_vocab_size
    ]
    logits.masked_fill_(mask, torch.finfo(logits.dtype).min)


@torch.no_grad()
def masked_normalization(
    x: torch.Tensor,
    mask: Optional[torch.BoolTensor] = None,
    dim=None,
    inplace=False,
    unbiased=False,
    eps=1e-5,
    high_precision=True,
):
    """Normalize x with a mask. Typically used in advantage normalization.

    Args:
        x (torch.Tensor):
            Tensor to be normalized.
        mask (torch.Tensor, optional):
            A mask with the same shape as x. Defaults to None.
        dim (int or tuple of ints, optional):
            Dimensions to be normalized. Defaults to None.
        inplace (bool, optional):
            Whether to perform in-place operation. Defaults to False.
        eps (torch.Tensor, optional):
            Minimal denominator. Defaults to 1e-5.

    Returns:
        torch.Tensor:
            Normalized x, with the same shape as x.
    """
    dtype = torch.float64 if high_precision else torch.float32
    x = x.to(dtype)
    if not inplace:
        x = x.clone()
    if dim is None:
        dim = tuple(range(len(x.shape)))
    if mask is None:
        factor = torch.tensor(
            np.prod([x.shape[d] for d in dim]), dtype=dtype, device=x.device
        )
    else:
        mask = mask.to(dtype)
        assert len(mask.shape) == len(x.shape), (mask.shape, x.shape, dim)
        for i in range(len(x.shape)):
            if i in dim:
                assert mask.shape[i] == x.shape[i], (mask.shape, x.shape, dim)
            else:
                assert mask.shape[i] == 1, (mask.shape, x.shape, dim)
        x = x * mask
        factor = mask.sum(dim, keepdim=True)
    x_sum = x.sum(dim=dim, keepdim=True)
    x_sum_sq = x.square().sum(dim=dim, keepdim=True)
    if dist.is_initialized():
        dist.all_reduce(
            factor, op=dist.ReduceOp.SUM, group=constants.data_parallel_group()
        )
        dist.all_reduce(
            x_sum, op=dist.ReduceOp.SUM, group=constants.data_parallel_group()
        )
        dist.all_reduce(
            x_sum_sq,
            op=dist.ReduceOp.SUM,
            group=constants.data_parallel_group(),
        )
    mean = x_sum / factor
    meansq = x_sum_sq / factor
    var = meansq - mean**2
    if unbiased:
        var *= factor / (factor - 1)
    return ((x - mean) / (var.sqrt() + eps)).float()


def get_eos_indices(
    input_ids: torch.LongTensor,
    tokenizer: transformers.PreTrainedTokenizerFast,
) -> Tuple[torch.LongTensor, torch.FloatTensor]:
    if torch.any(input_ids[:, 0] == tokenizer.eos_token_id):
        indices = (
            (input_ids[:, 0] == tokenizer.eos_token_id).nonzero().flatten()
        )
        bad_input_ids = input_ids[indices]
        bad_strs = tokenizer.batch_decode(
            bad_input_ids,
            skip_special_tokens=True,
            clean_up_tokenization_spaces=True,
        )
        raise RuntimeError(
            f"Generated sequence terminates unexpectedly early: {bad_strs}"
        )
    seq_len = input_ids.shape[1]
    eos_mask = (input_ids == tokenizer.eos_token_id).float()
    seq_no_eos_mask = (eos_mask.sum(1) == 0).float()
    eos_indices = eos_mask.argmax(1)
    eos_indices = (
        eos_indices * (1 - seq_no_eos_mask) + seq_no_eos_mask * (seq_len - 1)
    ).long()
    return eos_indices, seq_no_eos_mask


def torch_attn_func(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    causal: bool,
    cu_seqlens_q: torch.IntTensor,
    max_seqlen_q: int,
    cu_seqlens_k: torch.IntTensor,
    max_seqlen_k: int,
    dropout_p: float,
    softmax_scale: float,
    upcast_unscale: float = 1.0,
) -> torch.Tensor:
    """PyTorch implementation of the attention function with a flash-attn-like realhf.api.

    We use this function to compare the output of our model and huggingface models.
    Flash-attn/float16/CUDAkernels will all more or less suffer from float point errors.
    We call this function with float32 and CPU to get the "ground truth" output.

    Args:
        q (torch.Tensor): Shape [total_seqlen, #q, head_dim].
        k (torch.Tensor): Shape [total_seqlen, #kv, head_dim].
        v (torch.Tensor): Shape [total_seqlen, #kv, head_dim].
        causal (bool): .
        dropout_p (float): .
        softmax_scale (float): .
        upcast_unscale (float, optional): Scale factor when upcastin attention scores.
            Defaults to 1.0.

    Returns:
        torch.Tensor: Attention score. Shape [bs, seqlen, #q, head_dim].
    """
    nq = q.shape[-2]
    nkv = k.shape[-2]
    n_rep = q.shape[-2] // k.shape[-2]
    bsz = cu_seqlens_q.shape[0] - 1
    # repeat k/v heads if n_kv_heads < n_heads
    k = repeat_kv(k, n_rep)  # (total_seqlen, nq, head_dim)
    v = repeat_kv(v, n_rep)  # (total_seqlen, nq, head_dim)

    input_lens_k = cu_seqlens_k[1:] - cu_seqlens_k[:-1]
    attention_mask_k = torch.arange(
        max_seqlen_k, dtype=torch.long, device="cpu"
    ).unsqueeze(0) < input_lens_k.unsqueeze(1)
    _, _pad_indices_k, _, _ = unpad_input(attention_mask_k, attention_mask_k)

    input_lens_q = cu_seqlens_q[1:] - cu_seqlens_q[:-1]
    attention_mask_q = torch.arange(
        max_seqlen_q, dtype=torch.long, device="cpu"
    ).unsqueeze(0) < input_lens_q.unsqueeze(1)
    _, _pad_indices_q, _, _ = unpad_input(attention_mask_q, attention_mask_q)

    q = pad_input(q, _pad_indices_q, bsz, max_seqlen_q)
    k = pad_input(k, _pad_indices_k, bsz, max_seqlen_k)
    v = pad_input(v, _pad_indices_k, bsz, max_seqlen_k)

    q = q.transpose(1, 2)  # (bs, nq, seqlen, head_dim)
    k = k.transpose(1, 2)
    v = v.transpose(1, 2)
    scores = torch.matmul(q, k.transpose(2, 3)) * softmax_scale

    mask = (
        attention_mask_k.unsqueeze(1)
        .unsqueeze(1)
        .repeat(1, nq, max_seqlen_q, 1)
    )  # [bs, nq, seqlen, seqlen]
    if causal:
        _ms = max(max_seqlen_q, max_seqlen_k)
        causal_mask = torch.tril(
            torch.ones(_ms, _ms, device=q.device, dtype=torch.bool)
        )[-max_seqlen_q:, -max_seqlen_k:]
        mask = mask & causal_mask

    # if mask_softmax:
    scores = upcast_masked_softmax(
        scores,
        mask,
        mask_value=torch.full(
            [],
            torch.finfo(torch.float32).min,
            device=scores.device,
            dtype=torch.float32,
        ),
        scale=upcast_unscale,
        softmax_dtype=torch.float32,
    )
    # else:
    #     scores = upcast_softmax(scores, scale=upcast_unscale, softmax_dtype=torch.float32)
    scores = torch.nn.functional.dropout(scores, p=dropout_p)
    scores = scores.to(q.dtype)
    output = torch.matmul(scores, v)  # (bs, nq, seqlen, head_dim)
    output = output.transpose(1, 2).contiguous()

    output = unpad_input(output, attention_mask_q)[0]
    return output


def rotate_half(x: torch.HalfTensor, interleaved: bool = False):
    if not interleaved:
        x1, x2 = x.chunk(2, dim=-1)
        return torch.cat((-x2, x1), dim=-1)
    else:
        x1, x2 = x[..., ::2], x[..., 1::2]
        # return rearrange(torch.stack((-x2, x1), dim=-1), "... d two -> ... (d two)", two=2)
        return torch.stack((-x2, x1), dim=-1).flatten(start_dim=-2)


@torch.no_grad()
@torch.jit.script
def compute_varlen_position_indices(
    total_seqlen: int,
    cu_seqlens: torch.IntTensor,
    seqlen_offsets: Optional[torch.IntTensor] = None,
) -> torch.IntTensor:
    indexing_t = torch.arange(
        total_seqlen, dtype=torch.long, device=cu_seqlens.device
    ).unsqueeze_(0)
    indexing_t = (cu_seqlens[:-1].unsqueeze(1) <= indexing_t) & (
        indexing_t < cu_seqlens[1:].unsqueeze(1)
    )
    indices = indexing_t.cumsum(1) - 1
    if seqlen_offsets is not None:
        indices += seqlen_offsets.unsqueeze(1)
    return torch.where(indexing_t, indices, 0).sum(0)


# @torch.jit.script
def apply_rotary_varlen(
    x: torch.HalfTensor,
    cos: torch.HalfTensor,
    sin: torch.HalfTensor,
    cu_seqlens: torch.IntTensor,
    interleaved: bool,
    seqlen_offsets: Optional[torch.IntTensor] = None,
    rotary_indices: Optional[torch.LongTensor] = None,
) -> Tuple[torch.HalfTensor, torch.LongTensor]:
    if rotary_indices is None:
        rotary_indices = compute_varlen_position_indices(
            x.shape[0], cu_seqlens, seqlen_offsets
        )

    ro_dim = cos.shape[-1] * 2
    assert ro_dim <= x.shape[-1]
    cos = cos[rotary_indices]
    sin = sin[rotary_indices]
    if not interleaved:
        cos = cos[:, None, None, :].repeat(1, 1, 2, 1).flatten(start_dim=-2)
        sin = sin[:, None, None, :].repeat(1, 1, 2, 1).flatten(start_dim=-2)
    else:
        cos = cos[:, None, :, None].repeat(1, 1, 1, 2).flatten(start_dim=-2)
        sin = sin[:, None, :, None].repeat(1, 1, 1, 2).flatten(start_dim=-2)

    # cos = repeat(cos, "... d -> ... 1 (2 d)" if not interleaved else "... d -> ... 1 (d 2)")
    # sin = repeat(sin, "... d -> ... 1 (2 d)" if not interleaved else "... d -> ... 1 (d 2)")
    return torch.cat(
        [
            x[..., :ro_dim] * cos
            + rotate_half(x[..., :ro_dim], interleaved) * sin,
            x[..., ro_dim:],
        ],
        dim=-1,
    )


def apply_rotary(
    x: torch.HalfTensor,
    cos: torch.HalfTensor,
    sin: torch.HalfTensor,
    interleaved: bool = False,
):
    """
    x: (batch_size, seqlen, nheads, headdim)
    cos, sin: (seqlen, rotary_dim / 2) or (batch_size, seqlen, rotary_dim / 2)
    """
    ro_dim = cos.shape[-1] * 2
    assert ro_dim <= x.shape[-1]
    if not interleaved:
        cos = cos[:, None, None, :].repeat(1, 1, 2, 1).flatten(start_dim=-2)
        sin = sin[:, None, None, :].repeat(1, 1, 2, 1).flatten(start_dim=-2)
    else:
        cos = cos[:, None, :, None].repeat(1, 1, 1, 2).flatten(start_dim=-2)
        sin = sin[:, None, :, None].repeat(1, 1, 1, 2).flatten(start_dim=-2)
    return torch.cat(
        [
            x[..., :ro_dim] * cos
            + rotate_half(x[..., :ro_dim], interleaved) * sin,
            x[..., ro_dim:],
        ],
        dim=-1,
    )



from asyncio.base_events import _run_until_complete_cb
import asyncio
import dataclasses
import sys
import threading


@dataclasses.dataclass
class AsyncRunUntilCompleteContext:
    loop: asyncio.BaseEventLoop
    future: asyncio.Future
    new_task: bool


def setup_run_until_complete(
    loop: asyncio.BaseEventLoop,
    future: asyncio.Future,
) -> AsyncRunUntilCompleteContext:
    loop._check_closed()
    loop._check_running()

    new_task = not asyncio.futures.isfuture(future)
    future = asyncio.tasks.ensure_future(future, loop=loop)
    if new_task:
        # An exception is raised if the future didn't complete, so there
        # is no need to log the "destroy pending task" message
        future._log_destroy_pending = False

    future.add_done_callback(_run_until_complete_cb)

    # set up run forever
    loop._set_coroutine_origin_tracking(loop._debug)

    loop._old_agen_hooks = sys.get_asyncgen_hooks()
    loop._thread_id = threading.get_ident()
    sys.set_asyncgen_hooks(
        firstiter=loop._asyncgen_firstiter_hook,
        finalizer=loop._asyncgen_finalizer_hook,
    )
    asyncio.events._set_running_loop(loop)
    return AsyncRunUntilCompleteContext(
        loop=loop, future=future, new_task=new_task
    )


def teardown_run_util_complete(ctx: AsyncRunUntilCompleteContext):
    ctx.loop._stopping = False
    ctx.loop._thread_id = None
    asyncio.events._set_running_loop(None)
    ctx.loop._set_coroutine_origin_tracking(False)
    # Restore any pre-existing async generator hooks.
    if ctx.loop._old_agen_hooks is not None:
        sys.set_asyncgen_hooks(*ctx.loop._old_agen_hooks)
        ctx.loop._old_agen_hooks = None

    ctx.future.remove_done_callback(_run_until_complete_cb)

    if not ctx.future.done():
        raise RuntimeError("Event loop stopped before Future completed.")


def raise_asyncio_exception(
    ctx: AsyncRunUntilCompleteContext, raise_error: bool = True
):
    if ctx.new_task and ctx.future.done() and not ctx.future.cancelled():
        # The coroutine raised a BaseException. Consume the exception
        # to not log a warning, the caller doesn't have access to the
        # local task.
        ctx.future.exception()

    try:
        teardown_run_util_complete(ctx)
    except RuntimeError as e:
        if raise_error:
            raise e

    if raise_error:
        raise



from typing import *
import dataclasses
import json
import os
import time

import torch
import torch.distributed as dist
import transformers

from realhf.api.core import model_api
from realhf.base import constants, logging
from realhf.base.saveload_utils import split_state_dict_into_shards
from realhf.impl.model.nn.real_llm_api import ReaLModel
from realhf.impl.model.nn.real_llm_parallel import (
    mp_merge_key,
    mp_partition_real_model_state_dict,
)

logger = logging.getLogger("HF Registry")


@dataclasses.dataclass
class HFModelRegistry:
    name: str
    hf_cls_name: str
    config_from_hf_converter: Callable[
        [transformers.PretrainedConfig], model_api.ReaLModelConfig
    ]
    config_to_hf_converter: Callable[
        [model_api.ReaLModelConfig], transformers.PretrainedConfig
    ]
    sd_from_hf_converter: Callable[[Dict, model_api.ReaLModelConfig], Dict]
    sd_to_hf_converter: Callable[[Dict, model_api.ReaLModelConfig], Dict]
    embedding_param_names: Callable[[model_api.ReaLModelConfig], List[str]]
    tblock_param_names: Callable[[model_api.ReaLModelConfig, int], List[str]]
    head_param_names: Callable[[model_api.ReaLModelConfig], List[str]]

    def config_from_hf(
        self,
        hf_config: Optional[transformers.PretrainedConfig] = None,
        model_path: Optional[str] = None,
        is_critic: bool = False,
    ) -> model_api.ReaLModelConfig:
        if hf_config is None:
            hf_config = transformers.AutoConfig.from_pretrained(
                os.path.join(model_path, "config.json")
            )
        config = self.config_from_hf_converter(hf_config)
        config.is_critic = is_critic
        return config

    def config_to_hf(
        self, real_config: model_api.ReaLModelConfig
    ) -> transformers.PretrainedConfig:
        return self.config_to_hf_converter(real_config)

    def load(
        self,
        model: ReaLModel,
        load_dir: str,
        init_critic_from_actor: bool = False,
    ):
        # NOTE: moving this upwards will result in circular import

        tik = time.perf_counter()
        with open(os.path.join(load_dir, "config.json"), "r") as f:
            hf_config = json.load(f)
        if "architectures" in hf_config:
            assert (
                self.hf_cls_name == hf_config["architectures"][0]
            ), f"{self.hf_cls_name} != {hf_config['architectures'][0]}"

        layer_indices = range(model.layer_idx_start, model.layer_idx_end)

        required_hf_sd_names = []
        for lidx in layer_indices:
            if lidx == 0:
                required_hf_sd_names += self.embedding_param_names(model.config)
            elif lidx == model.config.n_layers + 1:
                required_hf_sd_names += self.head_param_names(model.config)
            else:
                required_hf_sd_names += self.tblock_param_names(
                    model.config, lidx - 1
                )

        if os.path.exists(
            os.path.join(load_dir, "pytorch_model.bin.index.json")
        ):
            with open(
                os.path.join(load_dir, "pytorch_model.bin.index.json"), "r"
            ) as f:
                hf_sd_mapping = json.load(f)["weight_map"]
            files_to_load = set(
                hf_sd_mapping[name] for name in required_hf_sd_names
            )
        else:
            files_to_load = ["pytorch_model.bin"]
        setup_time = time.perf_counter() - tik

        load_times, partition_times = [], []
        state_dict = {}
        for fn in files_to_load:
            load_tik = time.perf_counter()
            # set map_location to be CPU is a little bit faster
            sd = torch.load(os.path.join(load_dir, fn), map_location="cpu")
            partition_tik = time.perf_counter()
            sd = {k: v for k, v in sd.items() if k in required_hf_sd_names}
            sd = self.sd_from_hf_converter(sd, model.config)
            psd = mp_partition_real_model_state_dict(
                sd,
                model.config,
                constants.model_parallel_world_size(),
                constants.model_parallel_rank(),
            )
            state_dict.update(psd)
            load_times.append(partition_tik - load_tik)
            partition_times.append(time.perf_counter() - partition_tik)

        copy_tik = time.perf_counter()
        if (
            init_critic_from_actor
            and f"{model.config.n_layers + 1}.weight" in state_dict
        ):
            state_dict.pop(f"{model.config.n_layers + 1}.weight")
            model.load_state_dict(state_dict, strict=False)
        else:
            try:
                model.load_state_dict(state_dict, strict=True)
            except Exception as e:
                logger.error(
                    f"Loading state dict with strict=True failed. "
                    f"Have you set init_critic_from_actor=True "
                    f"in the model config if you are initializing "
                    f"a critic model from a regular LLM? Err: {e}"
                )

        # Some logging info
        copy_time = time.perf_counter() - copy_tik
        load_times = "[" + ", ".join(f"{t:.2f}" for t in load_times) + "]"
        partition_times = (
            "[" + ", ".join(f"{t:.2f}" for t in partition_times) + "]"
        )
        if os.getenv("REAL_LOG_LOAD_TIME", None) == "1":
            logger.info(
                f"Loading from HuggingFace Model setup time cost={setup_time:.2f}s, load time cost={load_times}, "
                f"partition time cost={partition_times}, copy time cost={copy_time:.2f}s"
            )
        return model

    def save(
        self,
        model: ReaLModel,
        tokenizer: transformers.PreTrainedTokenizer,
        save_dir: str,
    ):
        tik = time.perf_counter()
        os.makedirs(save_dir, exist_ok=True)

        dp_rank = constants.data_parallel_rank()
        pp_rank = constants.pipe_parallel_rank()
        mp_rank = constants.model_parallel_rank()
        mp_size = constants.model_parallel_world_size()
        pp_size = constants.pipe_parallel_world_size()
        dp_size = constants.data_parallel_world_size()

        # We will gather parameters across the model parallel group,
        # and save parameters to separate shards across the pipeline parallel group.

        # To decrease the size of each saved file, we split the file
        # of each pipeline stage into smaller shards.
        approx_param_size = (
            sum(
                v.numel() * v.element_size()
                for v in model.state_dict().values()
            )
            * mp_size
        )

        # By default a shard is at most 1GB. A small size enables parallel saving during training.
        max_shard_size_byte = int(
            os.getenv("REAL_SAVE_MAX_SHARD_SIZE_BYTE", int(1e9))
        )
        n_shards_this_stage = (
            approx_param_size + max_shard_size_byte - 1
        ) // max_shard_size_byte
        if approx_param_size <= 0 or n_shards_this_stage <= 0:
            raise ValueError(
                f"Invalid param_size={approx_param_size}, n_shards_this_stage={n_shards_this_stage}. "
                "Have you instantiated the model?"
            )

        n_shards_this_stage = torch.tensor(
            n_shards_this_stage, dtype=torch.int32, device="cuda"
        )
        pp_stage_n_shards = [
            torch.zeros_like(n_shards_this_stage) for _ in range(pp_size)
        ]
        dist.all_gather(
            pp_stage_n_shards,
            n_shards_this_stage,
            group=constants.pipe_parallel_group(),
        )
        pp_stage_n_shards = [int(n.item()) for n in pp_stage_n_shards]
        assert all(x >= 1 for x in pp_stage_n_shards)

        t1 = time.perf_counter()

        # Gather parameters across the model parallel group.
        sd = model.state_dict()
        cpu_sd = {}
        for k, v in sd.items():
            if (
                "k_attn" in k or "v_attn" in k
            ) and model.config.n_kv_heads % mp_size != 0:
                gathered = v
            else:
                gather_list = [torch.zeros_like(v) for _ in range(mp_size)]
                dist.all_gather(
                    gather_list, v, group=constants.model_parallel_group()
                )
                gathered = mp_merge_key(k, gather_list, model.config)
            cpu_sd[k] = gathered.cpu()

        t2 = time.perf_counter()

        hf_sd = self.sd_to_hf_converter(cpu_sd, model.config)
        hf_config = self.config_to_hf_converter(model.config)

        param_size = sum(
            [value.numel() * value.element_size() for value in hf_sd.values()]
        )
        param_size = torch.tensor(param_size, dtype=torch.int64, device="cuda")
        dist.all_reduce(
            param_size,
            op=dist.ReduceOp.SUM,
            group=constants.pipe_parallel_group(),
        )
        param_size = param_size.item()

        # Save tokenizer and huggingface model config.
        if pp_rank == 0 and dp_rank == 0 and mp_rank == 0:
            hf_config.save_pretrained(save_dir)
            tokenizer.save_pretrained(save_dir)

        # Dump parameters to disk.
        if len(pp_stage_n_shards) == 1 and pp_stage_n_shards[0] == 1:
            fn = "pytorch_model.bin"
            if pp_rank == 0 and dp_rank == 0 and mp_rank == 0:
                torch.save(hf_sd, os.path.join(save_dir, fn))
        else:
            output_fn = (
                "pytorch_model"
                + "-{shard:05d}"
                + f"-of-{sum(pp_stage_n_shards):05d}.bin"
            )

            n_shards = pp_stage_n_shards[pp_rank]
            shard_offset = sum(pp_stage_n_shards[:pp_rank])

            shards = split_state_dict_into_shards(hf_sd, n_shards)

            bin_index = {}
            bin_index["metadata"] = dict(total_size=param_size)
            bin_index["weight_map"] = {}
            weight_map = {}

            mesh_size = dp_size * mp_size
            mesh_idx = dp_rank * mp_size + mp_rank
            n_shards_per_gpu = (n_shards + mesh_size - 1) // mesh_size
            if mesh_idx < len(range(0, n_shards, n_shards_per_gpu)):
                s = list(range(0, n_shards, n_shards_per_gpu))[mesh_idx]
            else:
                s = n_shards

            for i, shard in enumerate(shards[s : s + n_shards_per_gpu]):
                shard_idx = shard_offset + i + s
                torch.save(
                    shard,
                    os.path.join(
                        save_dir, output_fn.format(shard=shard_idx + 1)
                    ),
                )

            for i, shard in enumerate(shards):
                shard_idx = shard_offset + i
                for k in shard:
                    weight_map[k] = output_fn.format(shard=shard_idx + 1)

            weight_map_list = [None for _ in range(pp_size)]
            dist.all_gather_object(
                weight_map_list,
                weight_map,
                group=constants.pipe_parallel_group(),
            )
            for wm in weight_map_list:
                bin_index["weight_map"].update(wm)

            if pp_rank == 0 and dp_rank == 0 and mp_rank == 0:
                with open(
                    os.path.join(save_dir, "pytorch_model.bin.index.json"), "w"
                ) as f:
                    json.dump(bin_index, f, indent=4)
        t3 = time.perf_counter()

        metadata_t = t1 - tik
        gather_cpu_t = t2 - t1
        dump_t = t3 - t2
        logger.info(
            f"Saving to HuggingFace Model metadata cost={metadata_t:.2f}s, gather/cpu copy cost={gather_cpu_t:.2f}s, "
            f"dump cost={dump_t:.2f}s"
        )



from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Union,
)
import collections
import dataclasses
import enum
import itertools

import numpy as np

from realhf.api.core.config import ModelBackend, ModelFamily, ModelName
import realhf.base.logging as logging
import realhf.base.namedarray as namedarray

logger = logging.getLogger("DataFlowGraph", "benchmark")


@dataclasses.dataclass
class OffloadHook:
    pass


@dataclasses.dataclass
class SyncParamHook:
    source: Optional[ModelName] = None
    target: Optional[ModelName] = None
    interval: int = 1


RPCHook = Union[OffloadHook, SyncParamHook]


@dataclasses.dataclass
class ModelInterface:
    type_: str
    args: Dict[str, Any] = dataclasses.field(default_factory=dict)

    def __eq__(self, other: "ModelInterface"):
        return self.type_ == other.type_ and self.args == other.args


class ModelInterfaceType(enum.Enum):
    GENERATE = "generate"
    TRAIN_STEP = "train_step"
    EVALUATE = "evaluate"
    INFERENCE = "inference"


@dataclasses.dataclass
class MFCDef:
    """This class defines a model function call (MFC) node in the dataflow graph.

    Dependencies are automatically resolved by the `build_graph` function
    according to the input/output data keys and the model name.

    :param model_name: The model identifier to be used by the node.
    :type model_name: ModelName
    :param model_type: The specification of the LLM, e.g., LLaMA-7B.
        Used by the profiler and search engine to produce an optimal execution plan.
    :type model_type: ModelFamily
    :param model_path: The path to the model file. Used to get the config.
    :type model_path: str
    :param interface_type: The interface type of this node, e.g., "generate".
    :type interface_type: ModelInterfaceType
    :param interface_impl: The actual interface implementation
        when running the this node.
    :type interface_impl: ModelInterface
    :param min_n_seqs: The minimum number of sequences to be processed in a batch.
    :type min_n_seqs: int
    :param max_n_seqs: The maximum number of sequences to be processed in a batch.
    :type max_n_seqs: int
    :param input_data: The data keys required by this node.
    :type input_data: List[str]
    :param input_key_remap: Remap input keys to let the interface implementation
        recognize them. Keys are ``input_data`` and values are the identifiers known to the
        interface.
    :type input_key_remap: Dict[str, str]
    :param output_data: The data keys generated by this node.
    :type output_data: List[str]
    :param output_key_remap: Remap output keys to let MFC recognize them.
        Keys are identifiers known to the interface and values are ``output_data``.
    :type output_key_remap: Dict[str, str]
    :param log_return_value: Whether to log the return value of the interface implementation.
    :type log_return_value: bool
    :param min_n_seqs_per_dp: The minimum number of sequences per data parallelism.
    :type min_n_seqs_per_dp: int
    :param balanced_dp: Whether to balance the data parallelism.
    :type balanced_dp: bool
    :param max_concurrent_calls: The maximum number of concurrent calls.
    :type max_concurrent_calls: int
    :param pre_hooks: The hooks to be executed before the interface implementation.
        Typically parameter reallocation. Should not be manually changed in
        using quickstart experiments.
    :type pre_hooks: List[RPCHook]
    :param post_hooks: The hooks to be executed after the interface implementation.
        Typically offload or parameter reallocation. Should not be manually changed in
        using quickstart experiments.
    :type post_hooks: List[RPCHook]
    """

    model_name: ModelName
    model_type: ModelFamily
    model_path: str
    interface_type: ModelInterfaceType
    interface_impl: ModelInterface

    # batch sizes
    min_n_seqs: int
    max_n_seqs: int

    input_data: List[str] = dataclasses.field(default_factory=lambda: [])
    input_key_remap: Dict[str, str] = dataclasses.field(
        default_factory=lambda: {}
    )
    output_data: List[str] = dataclasses.field(default_factory=lambda: [])
    output_key_remap: Dict[str, str] = dataclasses.field(
        default_factory=lambda: {}
    )
    log_return_value: bool = False

    min_n_seqs_per_dp: int = 1
    balanced_dp: bool = False

    max_concurrent_calls: int = 1

    # hooks
    pre_hooks: List[RPCHook] = dataclasses.field(default_factory=lambda: [])
    post_hooks: List[RPCHook] = dataclasses.field(default_factory=lambda: [])

    # The followings will be automatically filled.
    max_min_flow_seqs: int = 1

    parents: List[str] = dataclasses.field(default_factory=lambda: [])
    children: List[str] = dataclasses.field(default_factory=lambda: [])

    parent_rpcs: List["MFCDef"] = dataclasses.field(default_factory=lambda: [])
    children_rpcs: List["MFCDef"] = dataclasses.field(
        default_factory=lambda: []
    )

    # data key -> model name
    data_producers: Dict[str, ModelName] = dataclasses.field(
        default_factory=lambda: {}
    )
    # data key -> rpc names
    data2required_rpc_names: Dict[str, List[str]] = dataclasses.field(
        default_factory=lambda: {}
    )

    def __post_init__(self):
        if isinstance(self.model_name, str):
            self.model_name = ModelName(self.model_name, 0)
        assert isinstance(self.model_name, ModelName)
        if self.min_n_seqs > self.max_n_seqs:
            raise RuntimeError("Invalid min/max n_seqs.")
        if self.is_src and self.max_n_seqs > 1e4:
            raise RuntimeError(
                "The maximum batch size of the source node in the dataflow graph is too large. "
                f"The maximum number of sequences is {self.max_n_seqs} > budget {int(1e4)}. "
                "Please set a smaller value."
            )
        if "@" in self.model_name.role or "@" in self.interface_type.value:
            raise ValueError(
                f"Invalid model name or interface type: {self.model_name}, {self.interface_type}."
            )

    def __repr__(self):
        return f"MFCDef({self.model_name}, {self.interface_type})"

    @property
    def name(self):
        return f"{self.model_name}@{self.interface_type.value}"

    @property
    def is_src(self):
        return len(self.parents) == 0

    @property
    def is_dst(self):
        return len(self.children) == 0

    @property
    def is_dst_of_model_role(self):

        def _has_children_of_model_name(rpc: "MFCDef", model_name: ModelName):
            if rpc.is_dst:
                return False
            return any(
                [
                    r.model_name.role == model_name.role
                    or _has_children_of_model_name(r, model_name)
                    for r in rpc.children_rpcs
                ]
            )

        return not _has_children_of_model_name(self, self.model_name)

    def remap_input_keys(self, input_batch: Dict) -> namedarray.NamedArray:
        data = {}
        for k in self.input_data:
            if k not in self.input_key_remap:
                data[k] = input_batch[k]
            else:
                data[self.input_key_remap[k]] = input_batch[k]
        return namedarray.from_dict(data)

    def remap_output_keys(self, output_batch: Dict) -> namedarray.NamedArray:
        res_data = {}
        for k, v in output_batch.items():
            if k not in self.output_data:
                continue
            if k in self.output_key_remap:
                res_data[self.output_key_remap[k]] = v
            else:
                res_data[k] = v
        return namedarray.from_dict(res_data)


def build_graph(
    rpcs: List[MFCDef], verbose: bool = False
) -> Tuple[List[MFCDef], List[List[Tuple[str]]]]:
    # Resolve dependencies between model interfaces.
    children: List[List[str]] = [[] for _ in rpcs]
    parents: List[List[str]] = [[] for _ in rpcs]
    parent_rpcs: List[List[MFCDef]] = [[] for _ in rpcs]
    children_rpcs: List[List[MFCDef]] = [[] for _ in rpcs]
    edges: List[List[Tuple[str]]] = [[() for _ in rpcs] for _ in rpcs]

    required_data_entries: List[Tuple[str]] = [() for _ in rpcs]
    generated_data_entries: List[Tuple[str]] = [() for _ in rpcs]
    for i, rpc in enumerate(rpcs):
        required_data_entries[i] = (*required_data_entries[i], *rpc.input_data)
        generated_data_entries[i] = (
            *generated_data_entries[i],
            *[
                k if k not in rpc.output_key_remap else rpc.output_key_remap[k]
                for k in rpc.output_data
            ],
        )
    data_producers = {}
    for rpc, gd in zip(rpcs, generated_data_entries):
        for k in gd:
            data_producers[k] = rpc.model_name
    data2required_rpc_names = collections.defaultdict(list)
    for rpc, data_keys in zip(rpcs, required_data_entries):
        for k in data_keys:
            data2required_rpc_names[k].append(rpc.name)

    for i, rpc in enumerate(rpcs):
        for k in required_data_entries[i]:
            for j, parent_rpc in enumerate(rpcs):
                if parent_rpc.name == rpc.name:
                    continue
                if k in generated_data_entries[j]:
                    if parent_rpc.name not in parents[i]:
                        parents[i].append(parent_rpc.name)
                        parent_rpcs[i].append(parent_rpc)
                    if rpc.name not in children[j]:
                        children[j].append(rpc.name)
                        children_rpcs[j].append(rpc)
                    edges[i][j] = (*edges[i][j], k)
                    if verbose:
                        logger.info(
                            f"Dependency added: {rpc.name}"
                            f" <- {parent_rpc.name} "
                            f"because of data entry `{k}`."
                        )
    if verbose:
        for i, rpc in enumerate(rpcs):
            logger.info(
                f"Dependency: {rpc.name} <- { {x.name: deps for x, deps in zip(rpcs, edges[i]) if deps} }."
            )
    for rpc, p, c, pr, cr in zip(
        rpcs, parents, children, parent_rpcs, children_rpcs
    ):
        rpc.parents = p
        rpc.children = c
        rpc.parent_rpcs = pr
        rpc.children_rpcs = cr

    for rpc in rpcs:
        rpc.max_min_flow_seqs = max(
            [
                r.min_n_seqs
                for r in rpcs
                if r.model_name.role == rpc.model_name.role
            ]
        )
        rpc.data_producers = data_producers
        rpc.data2required_rpc_names = data2required_rpc_names

    # sanity check of hooks
    for rpc in rpcs:
        for h in itertools.chain(rpc.pre_hooks, rpc.post_hooks):
            assert isinstance(h, RPCHook), type(h)
            if isinstance(h, SyncParamHook):
                assert any(h.target == r.model_name for r in rpcs) or (
                    h.source == r.model_name for r in rpcs
                )
        for h in rpc.pre_hooks:
            if isinstance(h, OffloadHook):
                raise ValueError("Offload can only be post hooks!")
    return rpcs, edges



from typing import *
import collections

from realhf.api.core.config import ModelBackend
from realhf.api.core.dfg import (
    MFCDef,
    ModelInterfaceType,
    OffloadHook,
    SyncParamHook,
)
from realhf.api.core.system_api import ModelName
from realhf.api.quickstart.device_mesh import DeviceMesh, RPCAllocation
from realhf.api.quickstart.model import (
    get_real_model_config,
    ModelTrainEvalConfig,
    ParallelismConfig,
)
from realhf.base.topology import PipeModelDataParallelTopology


def get_topo(
    parallel: ParallelismConfig,
    gradient_checkpointing: bool,
    max_prompt_len: Optional[int] = None,
) -> PipeModelDataParallelTopology:
    return PipeModelDataParallelTopology(
        num_mp=parallel.model_parallel_size,
        num_pp=parallel.pipeline_parallel_size,
        num_dp=parallel.data_parallel_size,
        sequence_parallel=parallel.use_sequence_parallel,
        gradient_checkpointing=gradient_checkpointing,
        max_prompt_len=max_prompt_len,
    )


def get_world_size(parallel: ParallelismConfig) -> int:
    return (
        parallel.model_parallel_size
        * parallel.pipeline_parallel_size
        * parallel.data_parallel_size
    )


def make_train_backend_config(
    model_cfg: ModelTrainEvalConfig, parallel_cfg: ParallelismConfig
):
    if model_cfg.backend == "deepspeed":
        return ModelBackend(
            "deepspeed",
            args=dict(
                optimizer_name="adam",
                optimizer_config=dict(
                    lr=model_cfg.optimizer.lr,
                    weight_decay=model_cfg.optimizer.weight_decay,
                    eps=model_cfg.optimizer.eps,
                    betas=(
                        model_cfg.optimizer.beta1,
                        model_cfg.optimizer.beta2,
                    ),
                ),
                lr_scheduler_type=model_cfg.optimizer.lr_scheduler_type,
                warmup_steps_proportion=model_cfg.optimizer.warmup_steps_proportion,
                min_lr_ratio=model_cfg.optimizer.min_lr_ratio,
                zero_stage=(
                    model_cfg.zero_stage
                    if parallel_cfg.pipeline_parallel_size == 1
                    else min(model_cfg.zero_stage, 1)
                ),
                offload_optimizer_state=model_cfg.optimizer.offload,
                offload_param=model_cfg.offload,
                enable_bf16=model_cfg.enable_bf16,
                enable_fp16=model_cfg.enable_fp16,
            ),
        )
    elif model_cfg.backend == "megatron":
        if model_cfg.optimizer.offload or model_cfg.offload:
            raise ValueError("Offload is not supported in Megatron backend.")
        if model_cfg.zero_stage == 3:
            raise ValueError(
                "Zero stage 3 is not supported in Megatron backend."
            )
        return ModelBackend(
            "megatron",
            args=dict(
                optimizer_name="adam",
                optimizer_config=dict(
                    lr=model_cfg.optimizer.lr,
                    weight_decay=model_cfg.optimizer.weight_decay,
                    eps=model_cfg.optimizer.eps,
                    betas=(
                        model_cfg.optimizer.beta1,
                        model_cfg.optimizer.beta2,
                    ),
                ),
                lr_scheduler_type=model_cfg.optimizer.lr_scheduler_type,
                warmup_steps_proportion=model_cfg.optimizer.warmup_steps_proportion,
                min_lr_ratio=model_cfg.optimizer.min_lr_ratio,
                enable_bf16=model_cfg.enable_bf16,
                enable_fp16=model_cfg.enable_fp16,
                use_zero_optimization=model_cfg.zero_stage > 0,
                overlap_grad_reduce=model_cfg.zero_stage > 1,
                overlap_param_gather=model_cfg.zero_stage > 2,
            ),
        )
    else:
        raise NotImplementedError(
            f"Backend {model_cfg.backend} is not supported."
        )


def make_inf_backend_config(
    model_cfg: ModelTrainEvalConfig, parallel_cfg: ParallelismConfig
):
    return ModelBackend("inference")


def make_model_config(cfg: ModelTrainEvalConfig):
    return get_real_model_config(
        model_path=cfg.path,
        hf_model_family=cfg.type._class,
        is_critic=cfg.type.is_critic,
        init_critic_from_actor=cfg.init_critic_from_actor,
        dtype="bf16" if cfg.enable_bf16 else "fp16",
        lora=cfg.lora,
    )


def resolve_rpc_hooks(rpc_allocs: List[RPCAllocation]):
    from realhf.api.quickstart.model import parallelism_config_equal

    role_cnt = collections.defaultdict(int)
    for rpc_alloc in rpc_allocs:
        rpc = rpc_alloc.rpc
        parallel = rpc_alloc.parallel
        device_mesh = rpc_alloc.device_mesh
        # check param realloc hooks for train_step rpcs
        if rpc.interface_type == ModelInterfaceType.TRAIN_STEP:
            for other in rpc_allocs:
                if rpc.name == other.rpc.name:
                    continue
                if rpc.model_name.role == other.rpc.model_name.role and not (
                    parallelism_config_equal(parallel, other.parallel)
                    and device_mesh == other.device_mesh
                ):
                    other.rpc.model_name = ModelName(
                        rpc.model_name.role, role_cnt[rpc.model_name.role] + 1
                    )
                    role_cnt[rpc.model_name.role] += 1
                    other.rpc.pre_hooks.append(
                        SyncParamHook(source=rpc.model_name)
                    )
                    other.rpc.post_hooks.append(
                        SyncParamHook(target=rpc.model_name)
                    )

        # add offload hooks for inference rpcs
        if rpc.interface_type == ModelInterfaceType.INFERENCE:
            offload_flag = True
            # if there is training rpcs for the same model, can not offload
            for other in rpc_allocs:
                if (
                    other.rpc.model_name.role == rpc.model_name.role
                    and other.rpc.interface_type
                    == ModelInterfaceType.TRAIN_STEP
                ):
                    offload_flag = False
            if offload_flag:
                rpc.post_hooks.append(OffloadHook())

        print("resolve rpc hooks", rpc.name, rpc.model_name)



def import_profiler_registers():
    import realhf.search_engine.enumerate
    import realhf.search_engine.estimate
    import realhf.search_engine.layers
    import realhf.search_engine.param_realloc
    import realhf.search_engine.search
    import realhf.search_engine.utils



from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union
import collections
import copy
import dataclasses
import enum
import getpass
import itertools
import math
import os
import sys

from realhf.api.core.config import *
from realhf.base.cluster import spec as cluster_spec
from realhf.base.constants import (
    DATASET_CACHE_PATH,
    PYTORCH_KERNEL_CACHE_PATH,
    TORCH_EXTENSIONS_DIR,
    TRITON_CACHE_PATH,
)
import realhf.api.core.dfg as dfg
import realhf.base.topology as topology

_LLM_ENVVARS = {
    # "NCCL_P2P_DISABLE": "1",
    # "NCCL_IB_DISABLE": "1",
    "TRANSFORMERS_OFFLINE": "1",
    "PYTORCH_KERNEL_CACHE_PATH": PYTORCH_KERNEL_CACHE_PATH,
    "TRITON_CACHE_DIR": TRITON_CACHE_PATH,
    "TOKENIZERS_PARALLELISM": "true",
    "TORCH_EXTENSIONS_DIR": TORCH_EXTENSIONS_DIR,
    # "NCCL_DEBUG": "INFO",
    # "TORCH_DISTRIBUTED_DEBUG": "DETAIL",
    # "NCCL_SOCKET_IFNAME": "ibp71s0",
    # "GLOO_SOCKET_IFNAME": "ibp71s0",
    # "TORCH_USE_CUDA_DSA": "1",
    # "NCCL_IGNORE_DISABLED_P2P": "1",
    # "CUDA_LAUNCH_BLOCKING": "1",  # NOTE: CUDAGraph Capturing will not work if CUDA_LAUNCH_BLOCKING is set to 1.
    # "NCCL_COMM_BLOCKING": "1",  # NOTE: CUDAGraph Capturing will not work if NCCL_COMM_BLOCKING is set to 1.
    # "NCCL_BLOCKING_WAIT": "1",  # NOTE: CUDAGraph Capturing will not work if NCCL_BLOCKING_WAIT is set to 1.
    # "TORCH_SHOW_CPP_STACKTRACES": "1",
    "RAY_DEDUP_LOGS": "0",  # disable ray log deduplication
    "CUDA_DEVICE_MAX_CONNECTIONS": "1",
    "PYTHONUSERBASE": "/nonsense",
    # torch.distributed.all_reduce does not free the input tensor until
    # the synchronization point. This causes the memory usage to grow
    # as the number of all_reduce calls increases. This env var disables
    # this behavior.
    # Related issue:
    # https://discuss.pytorch.org/t/cuda-allocation-lifetime-for-inputs-to-distributed-all-reduce/191573
    "TORCH_NCCL_AVOID_RECORD_STREAMS": "1",
    # Whether to enable time mark to plot timelines.
    "REAL_CUDA_TMARK": "1",
}
for k, v in _LLM_ENVVARS.items():
    os.environ[k] = v

_LLM_GPU_IMAGE = cluster_spec.gpu_image
_LLM_CPU_IMAGE = cluster_spec.cpu_image


@dataclasses.dataclass
class Scheduling:
    cpu: int
    gpu: int
    mem: int
    gpu_type: str = "tesla"
    node_type: str = None
    nodelist: str = None
    exclude: str = None
    container_image: str = _LLM_CPU_IMAGE
    env_vars: Dict[str, str] = dataclasses.field(
        default_factory=lambda: _LLM_ENVVARS
    )
    # time utils from "https://slurm.schedmd.com/sbatch.html"
    time_limit: Optional[str] = None  # see  "--time" option for format
    begin: Optional[str] = None  # see "--begin" option for format
    deadline: Optional[str] = None  # see "--deadline" option for format

    @staticmethod
    def master_worker_default(**kwargs):
        return Scheduling(
            **{
                "cpu": 16,
                "mem": 20 * 1024,
                "gpu": 0,
                "container_image": _LLM_CPU_IMAGE,
                **kwargs,
            }
        )

    @staticmethod
    def model_worker_default(**kwargs):
        return Scheduling(
            **{
                "cpu": 2,
                "gpu": 1,
                "mem": 60 * 1024,
                "container_image": _LLM_GPU_IMAGE,
                **kwargs,
            }
        )


@dataclasses.dataclass
class WorkerInformation:
    """The basic information of an worker. To improve config readability, the experiment starter will fill the
    fields, instead of letting the users do so in experiment configs.
    """

    experiment_name: str = ""
    trial_name: str = (
        ""  # Name of the trial of the experiment; e.g. "{USER}-0".
    )
    worker_type: str = ""  # E.g. "policy", "actor", or "trainer".
    worker_index: int = (
        -1
    )  # The index of the worker of the specific type, starting from 0.
    worker_count: int = (
        0  # Total number of workers; hence, 0 <= worker_index < worker_count.
    )
    worker_tag: Optional[str] = (
        None  # For actor and policy worker, can be "training" or "evaluation".
    )
    host_key: Optional[str] = (
        None  # Worker will update and keep this key alive.
    )
    watch_keys: Union[str, List[str]] = (
        None  # Worker will exit if all of the watching keys are gone.
    )
    wandb_entity: Optional[str] = (
        None  # wandb_{config} are optional. They overwrite system wandb_configuration.
    )
    wandb_project: Optional[str] = None
    wandb_job_type: Optional[str] = None
    wandb_group: Optional[str] = None
    wandb_name: Optional[str] = None
    wandb_config: Optional[Dict] = None
    log_wandb: Optional[bool] = None

    def system_setup(
        self,
        experiment_name,
        trial_name,
        worker_type,
        worker_index,
        worker_count,
    ):
        """Setup system related worker information, while leaving the rest untouched."""
        self.experiment_name = experiment_name
        self.trial_name = trial_name
        self.worker_type = worker_type
        self.worker_index = worker_index
        self.worker_count = worker_count


@dataclasses.dataclass
class ModelWorker:
    seed: int
    shards: List[StandaloneModelShard]
    # dataset, for source model workers
    tokenizer_name_or_path: Optional[str] = None
    datasets: Optional[List[Union[str, Dataset]]] = None
    dataloader: Union[str, DataLoader] = "packed"
    use_dataset_cache: bool = False
    dataset_cahce_root: str = DATASET_CACHE_PATH
    # cuda & cudnn config
    cudnn_benchmark: bool = False
    cudnn_deterministic: bool = False
    cuda_cache_cleanliness: bool = True
    cuda_cache_clear_freq: int = 10
    # model_topos and worker_info will be configured automatically
    model_rpcs: List[dfg.MFCDef] = None
    model_topos: Dict[ModelName, topology.PipeModelDataParallelTopology] = None
    msid2mwid: Dict[ModelShardID, int] = None
    data_transfer_pairs: List[Tuple[str, str]] = None
    sync_param_pairs: List[Tuple[str, str]] = None
    worker_info: Optional[WorkerInformation] = None

    def __post_init__(self):
        model_names = [s.id.model_name for s in self.shards]
        if len(set(model_names)) != len(model_names):
            raise ValueError(
                f"ModelWorker cannot have multiple shards of the same model name: {model_names}."
            )


@dataclasses.dataclass
class ExperimentSaveEvalControl:
    total_train_epochs: int = 1
    # save control
    save_frequency_epochs: Optional[int] = None
    save_frequency_steps: Optional[int] = None
    save_frequency_seconds: Optional[int] = None
    # eval control
    eval_frequency_epochs: Optional[int] = None
    eval_frequency_steps: Optional[int] = None
    eval_frequency_seconds: Optional[int] = None
    # benchmark
    benchmark_steps: Optional[int] = None


@dataclasses.dataclass
class MasterWorker:
    exp_ctrl: ExperimentSaveEvalControl
    # main components
    model_rpcs: List[dfg.MFCDef]
    n_model_workers: int
    model_topos: Dict[ModelName, topology.PipeModelDataParallelTopology]
    msid2mwid: Dict[ModelShardID, int] = None
    data_transfer_pairs: List[Tuple[str, str]] = None
    sync_param_pairs: List[Tuple[str, str]] = None
    worker_info: Optional[WorkerInformation] = None


@dataclasses.dataclass
class TasksGroup:
    count: int
    scheduling: Scheduling


@dataclasses.dataclass
class ExperimentScheduling:
    model_worker: Union[List[TasksGroup], TasksGroup] = dataclasses.field(
        default_factory=list
    )
    master_worker: Union[List[TasksGroup], TasksGroup] = dataclasses.field(
        default_factory=list
    )
    controller_image: str = _LLM_CPU_IMAGE


@dataclasses.dataclass
class ExperimentConfig:
    exp_ctrl: ExperimentSaveEvalControl
    # dataflow
    model_rpcs: List[dfg.MFCDef]
    model_worker: List[ModelWorker] = dataclasses.field(default_factory=list)
    # master_worker will be set automatically
    master_worker: Optional[List[MasterWorker]] = None
    config: Optional[Any] = None

    def __post_init__(self):
        assert self.master_worker is None
        model_names = set()
        for w in self.model_worker:
            model_names = model_names.union([s.id.model_name for s in w.shards])
        model_names = sorted(list(model_names))

        ############### Sanity check of model names ###############
        _roles = set(mn.role for mn in model_names)
        _replica_ids = {
            _role: sorted(
                [mn.replica_id for mn in model_names if mn.role == _role]
            )
            for _role in _roles
        }
        for v in _replica_ids.values():
            if list(sorted(v)) != list(range(len(v))):
                raise ValueError(
                    f"Model replica ids should be 0, 1, 2, ... for each role: {_replica_ids}."
                )
        ############### Sanity check of model names ###############

        model_topos: Dict[ModelName, topology.PipeModelDataParallelTopology] = (
            {}
        )
        model_configs: Dict[ModelName, Model] = {}
        for model_name in model_names:
            _this_mws = list(
                filter(
                    lambda mw: any(
                        x.id.model_name == model_name for x in mw.shards
                    ),
                    self.model_worker,
                )
            )
            all_shards: List[StandaloneModelShard] = [
                next(filter(lambda x: x.id.model_name == model_name, mw.shards))
                for mw in _this_mws
            ]
            # TODO: same topo, diff nodelist?
            # for k, v in model_topos.items():
            # if k.role == model_name.role and v == all_shards[0].id.topo:
            #     raise ValueError(f"If different RPCs have the same topology, "
            #                      f"they don't need to use multiple model names ({k}, {model_name}).")
            model_topos[model_name] = all_shards[0].id.topo
            model_configs[model_name] = all_shards[0].model

            ##### Sanity check of parallelism ranks. #####
            ranks = [s.id.parallelism_rank for s in all_shards]
            _topos = [s.id.topo for s in all_shards]
            if set(ranks) != set(list(range(len(_this_mws)))) or any(
                _t.world_size() != _topos[0].world_size() for _t in _topos
            ):
                raise ValueError(
                    f"Parallelism rank check failed: model name {model_name}, "
                    f"model shard ids={[s.id for s in all_shards]}."
                )
            ##### Sanity check of parallelism ranks. #####

        data_transfer_pairs: List[Tuple[ModelName, ModelName]] = []
        _rpc_nodes, edges = dfg.build_graph(self.model_rpcs, verbose=True)
        for i in range(len(self.model_rpcs)):
            for j in range(len(self.model_rpcs)):
                if len(edges[i][j]) > 0:
                    # NOTE: dependencies are reversed here
                    data_transfer_pairs.append(
                        (
                            self.model_rpcs[j].model_name,
                            self.model_rpcs[i].model_name,
                        )
                    )
                    # print(f"model_rpc j name {self.model_rpcs[j].name} i name {self.model_rpcs[i].name} "
                    #        "data transfer pair append {(self.model_rpcs[j].model_name, self.model_rpcs[i].model_name)} ")
        src_rpcs = [rpc for rpc in self.model_rpcs if rpc.is_src]
        data_src_rpc = src_rpcs[0]
        for r in src_rpcs[1:]:
            if (
                data_src_rpc.model_name,
                r.model_name,
            ) not in data_transfer_pairs:
                data_transfer_pairs.append(
                    (data_src_rpc.model_name, r.model_name)
                )
        data_transfer_pairs += [(mn, mn) for mn in model_names]

        sync_param_pairs: List[Tuple[ModelName, ModelName]] = []
        ######### sanity check of sync param hooks #########
        for rpc in self.model_rpcs:
            for hook in rpc.pre_hooks + rpc.post_hooks:
                if not isinstance(hook, dfg.SyncParamHook):
                    continue
                if (
                    hook.target is not None
                    and not (
                        model_configs[rpc.model_name].type_
                        == model_configs[hook.target].type_
                        == "real_model"
                    )
                ) or (
                    hook.source is not None
                    and not (
                        model_configs[rpc.model_name].type_
                        == model_configs[hook.source].type_
                        == "real_model"
                    )
                ):
                    raise ValueError(
                        "To synchronize parameters between two models, both models must be ReaLModel."
                    )
                other_model_name = (
                    hook.target if hook.target is not None else hook.source
                )
                other_topo = (
                    model_topos[hook.target]
                    if hook.target is not None
                    else model_topos[hook.source]
                )
                self_topo = model_topos[rpc.model_name]
                if (
                    self_topo.get_dim("model") % other_topo.get_dim("model")
                    != 0
                    and other_topo.get_dim("model") % self_topo.get_dim("model")
                    != 0
                ):
                    raise ValueError(
                        "To synchronize parameters between two models, "
                        "their model parallel size must be a multiple of each other."
                    )
                if rpc.model_name == other_model_name:
                    raise ValueError(
                        f"Cannot synchronize parameters within the same model "
                        f"(in {rpc}, {rpc.model_name} and {hook.target})."
                    )
                if hook.target is not None:
                    if not (rpc.model_name, hook.target) in sync_param_pairs:
                        sync_param_pairs.append((rpc.model_name, hook.target))
                else:
                    if not (hook.source, rpc.model_name) in sync_param_pairs:
                        sync_param_pairs.append((hook.source, rpc.model_name))
        param_senders = set([x[0] for x in sync_param_pairs])
        assert len(set(param_senders)) == len(param_senders)
        param_receivers = set([x[1] for x in sync_param_pairs])
        assert len(set(param_receivers)) == len(param_receivers)
        for a, b in sync_param_pairs:
            if (b, a) not in sync_param_pairs:
                raise ValueError(
                    "The current implementation of parameter synchronization "
                    f"will throw local parameters away, "
                    f"so it is necceary to do bidirectional synchronization. "
                    f"{(a,b)} found in sync param pairs but not {(b,a)}"
                )
        ######### sanity check of sync param hooks #########

        # Mark which shard of the same role should be instantiated.
        _model_is_trainable = collections.defaultdict(list)
        for rpc in self.model_rpcs:
            _model_is_trainable[rpc.model_name].append(
                rpc.interface_type == dfg.ModelInterfaceType.TRAIN_STEP
            )

        _model_is_trainable = {
            model_name: any(values)
            for model_name, values in _model_is_trainable.items()
        }

        _roles = set([rpc.model_name.role for rpc in _rpc_nodes])
        _role_cnt = {
            role: len(
                set(
                    [
                        rpc.model_name
                        for rpc in _rpc_nodes
                        if rpc.model_name.role == role
                    ]
                )
            )
            for role in _roles
        }
        model_names_to_instantiate = []
        for role in _roles:
            _trainable_this_role = [
                _model_is_trainable[ModelName(role, i)]
                for i in range(_role_cnt[role])
            ]
            if _role_cnt[role] == 1 or not any(_trainable_this_role):
                model_names_to_instantiate.append(ModelName(role, 0))
                continue
            if any(_trainable_this_role):
                assert sum(_trainable_this_role) == 1
                _trainable_idx = _trainable_this_role.index(True)
                model_names_to_instantiate.append(
                    ModelName(role, _trainable_idx)
                )
        for mw in self.model_worker:
            for s in mw.shards:
                s.should_instantiate = (
                    s.id.model_name in model_names_to_instantiate
                )

        msid2mwid = {}
        for i, mw in enumerate(self.model_worker):
            mw.model_topos = model_topos
            for m in mw.shards:
                msid2mwid[m.id] = i
        for m in self.model_worker:
            m.msid2mwid = msid2mwid
            m.data_transfer_pairs = data_transfer_pairs
            m.sync_param_pairs = sync_param_pairs

        for m in self.model_worker:
            m.model_rpcs = self.model_rpcs

        self.master_worker = [
            MasterWorker(
                exp_ctrl=self.exp_ctrl,
                model_topos=model_topos,
                model_rpcs=self.model_rpcs,
                n_model_workers=len(self.model_worker),
                msid2mwid=msid2mwid,
                sync_param_pairs=sync_param_pairs,
                data_transfer_pairs=data_transfer_pairs,
            )
        ]

    def set_worker_information(self, experiment_name, trial_name):
        if len(self.model_worker) > 0:
            assert len(self.master_worker) == 1

        for worker_type, workers in [
            ("model_worker", self.model_worker),
            ("master_worker", self.master_worker),
        ]:
            if len(workers) == 0:
                continue
            for i, worker in enumerate(workers):
                system_worker_info = dict(
                    experiment_name=experiment_name,
                    trial_name=trial_name,
                    worker_type=worker_type,
                    worker_index=i,
                    worker_count=len(workers),
                )
                if worker.worker_info is not None:
                    worker.worker_info.system_setup(**system_worker_info)
                else:
                    worker.worker_info = WorkerInformation(**system_worker_info)


class Experiment:
    """Base class for defining the procedure of an experiment."""

    def scheduling_setup(self, config) -> ExperimentScheduling:
        """Returns the Scheduling of all workers."""
        raise NotImplementedError()

    def initial_setup(self, config) -> ExperimentConfig:
        """Returns a list of workers to create when a trial of the experiment is initialized."""
        raise NotImplementedError()


def dump_config_to_yaml(config, file):
    import yaml

    with open(file, "w") as f:
        yaml.dump(dataclass_to_dict(config), f)


def load_config_from_yaml(file):
    import yaml

    with open(file, "r") as f:
        return config_to_dataclass(yaml.safe_load(f))


def dataclass_to_dict(dc):
    if isinstance(dc, (str, int, float)) or dc is None:
        pass
    elif isinstance(dc, enum.Enum):
        dc = dc.value
    elif isinstance(dc, (list, tuple)):
        dc = [dataclass_to_dict(d) for d in dc]
    elif isinstance(dc, dict):
        dc = {k: dataclass_to_dict(v) for k, v in dc.items()}
    elif dataclasses.is_dataclass(dc):
        root_name = dc.__class__.__name__
        dc = dict(
            config_class=root_name,
            config_value={
                k.name: dataclass_to_dict(getattr(dc, k.name))
                for k in dataclasses.fields(dc)
            },
        )
    else:
        raise f"{dc} of type {type(dc)} cannot be parse to dict."
    return dc


def config_to_dataclass(config: Union[List, Dict]):
    if isinstance(config, (list, tuple)):
        return [config_to_dataclass(c) for c in config]
    elif isinstance(config, dict):
        if "config_class" in config.keys():
            return getattr(sys.modules[__name__], config["config_class"])(
                **{
                    k: config_to_dataclass(v)
                    for k, v in config["config_value"].items()
                }
            )
        else:
            return config
    elif isinstance(config, (str, int, float)) or config is None:
        return config
    else:
        raise NotImplementedError(config)


ALL_EXPERIMENT_CLASSES = {}


def register_experiment(name, cls):
    assert name not in ALL_EXPERIMENT_CLASSES
    ALL_EXPERIMENT_CLASSES[name] = cls


def make_experiment(name) -> Experiment:
    cls = ALL_EXPERIMENT_CLASSES[name]
    return cls()
